"""
Interfaces for configuring enDAQ recorders.

Note: Wi-Fi configuration is done through the command interface. Changes
though the configuration interface are applied when the device next
resets or starts recording, while Wi-Fi changes take effect immediately.
This also keeps Wi-Fi access point passwords secret. Similarly, setting the
device's realtime clock is also done through the command interface, as it
also takes effect immediately.
"""

import errno
import logging
import os.path
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union, TYPE_CHECKING
import warnings

from ebmlite.core import loadSchema
from ebmlite.core import Document, Element, MasterElement
from idelib.dataset import Channel, SubChannel

from .exceptions import ConfigError, DeviceError, UnsupportedFeature
from . import legacy
from . import ui_defaults
from . import util

if TYPE_CHECKING:
    from .base import Recorder

logger = logging.getLogger('endaq.device')


# ===========================================================================
#
# ===========================================================================

# Default post-configuration message. Will (eventually) be in CONFIG.UI data.
_POST_CONFIG_MSG = ("When ready...\n"
                    "    1. Disconnect the recorder\n"
                    "    2. Mount to surface\n"
                    "    3. Press the recorder's primary button ")

# ===========================================================================
#
# ===========================================================================

class ConfigItem:
    """ A single configuration item/field, read from Config UI data, e.g., a
        device's ``CONFIG.UI`` file. It keeps track of the item's data type
        and handles conversion between internal and real-world units where
        applicable. It can also perform basic validation (minimum, maximum,
        specific valid options, etc.) if the ``CONFIG.UI`` data provides the
        required information.
    """

    # Mapping of field element names (exact match) to object attributes.
    ARGS = {"Label": "label",
            "ToolTip": "tooltip",
            "Units": "units",
            "DisplayFormat": "displayFormat",
            "ValueFormat": "valueFormat",
            "MaxLength": "maxLength",
            "EnumOption": "options"}

    # Mapping of *Field EBML IDs last 4 bits to value types.
    # For generating a default `vtype` for fields w/o a default *Value.
    TYPES = {0x00: "BooleanValue",
             0x01: "UIntValue",
             0x02: "IntValue",
             0x03: "FloatValue",
             0x04: "ASCIIValue",
             0x05: "TextValue",
             0x06: "BinaryValue",
             0x07: "UIntValue",  # enum, bitfield
             0x08: "BooleanValue"}  # group (e.g. a CheckGroup)

    # The low nybble of *Field EBML IDs represents their data type.
    DTYPES = {0x0: bool,
              0x1: int,
              0x2: int,
              0x3: float,
              0x4: str,
              0x5: str,
              0x6: bytes}

    # Names for configuration item types based on the high 8 bits of the
    # ConfigID, for generating labels for *Fields without Label elements
    TYPE_LABELS = {0x010000: 'Enable',
                   0x020000: 'Sample Rate',
                   0x030000: 'Low Trigger',
                   0x040000: 'High Trigger',
                   0x050000: 'Trigger Enable'}

    # Labels and tool tip strings for standard *Fields (usually special types)
    # that don't have `Label` and/or `ToolTip` elements.
    DEFAULT_LABELS = {
        0x0bff7f: ('UTC Offset', "The local timezone's offset from UTC time.")
    }

    # Default expression code objects for ValueFormat and DisplayFormat.
    # `noEffect` always returns the field's value unmodified (supplied as the
    # variable ``x``).
    noEffect = compile("x", "<ConfigItem.noEffect>", "eval")


    _tagsDisplay = compile("[t.strip() for t in x.split(',')]",
                           "<ConfigItem._tagDisplay>", "eval")

    _tagsValue = compile("','.join(str(x).strip()) if x else ''",
                         "<ConfigItem._tagValue>", "eval")

    @classmethod
    def _generateLabel(cls, configId: int) -> Union[str, None]:
        """ Helper method to create a 'label' string from a ConfigID (using
            the standard conventions) if the \\*Field element does not contain
            one. Mainly works for standard channel/subchannel-specific fields.
        """
        if configId in cls.DEFAULT_LABELS:
            return cls.DEFAULT_LABELS[configId][0]

        subch = (configId >> 8) & 0xFF
        ch = configId & 0xFF

        t = configId & 0xFF0000
        label = cls.TYPE_LABELS.get(t, None)

        if label is None:
            if ch == 0x7f:
                # Device-wide config item, no label
                return None
            # Config item type unknown
            label = "ConfigID 0x{:06x}".format(configId)

        if subch != 0xff:
            return "SubChannel {}.{} {}".format(ch, subch, label)
        return "Channel {} {}".format(ch, label)


    def __init__(self,
                 interface,
                 element: MasterElement,
                 data: Optional[dict] = None,
                 value: Optional[Any] = None):
        """
        `ConfigItem` is rarely (if ever) explicitly instantiated; instances
        are automatically created by a `ConfigInterface`, using CONFIG.UI
        data.

        :param interface: The 'parent' `ConfigInterface`.
        :param element: The raw CONFIG.UI `*Field` EBML element.
        :param data: The `element` contents, dumped as a dictionary. Avoids
            redundant dumping.
        :param value: The raw value (in native units) as read from a config
            file.
        """
        if data is None:
            data = element.dump()
        self.configId = data['ConfigID']

        # These attributes don't appear in ARGS
        self.interface = interface
        self.element = element
        self._default = None  # field default, in internal units
        self.vtype = None  # Type of *Value element in config data
        self.dtype = self.DTYPES.get(element.id & 0x0F, None)  # Value's Python data type

        # Internal/engineering unit conversion
        self.displayFormat = self.valueFormat = None
        self.gain = self.offset = None

        # Validation
        self.min = float('-inf')
        self.max = float('inf')
        self.maxLength = float('inf')
        self.options = []

        # For future use (if any)
        self.label = self.tooltip = self.units = None

        for k, v in data.items():
            if k in self.ARGS:
                setattr(self, self.ARGS[k], v)
            elif k.endswith('Value'):
                # Config item type determined by *Value element type
                # (if present; fallback behaviors below)
                self.vtype, self._default = k, v
            else:
                # Elements with data type as prefix
                for attr in ("Min", "Max", "Gain", "Offset"):
                    if k.endswith(attr):
                        setattr(self, attr.lower(), v)

        if self.options:
            # Choices for an EnumField or BitField
            self.options = self.parseOptions(self.options)

        if self.vtype is None:
            # No default *Value, and no type in an EnumOption;
            # fall back to type encoded in *Field EBML ID
            self.vtype = self.TYPES.get(self.element.id & 0x0f, None)

        if self.dtype is None and interface and self.vtype:
            # Data type not in dictionary of types (probably an *Enum);
            # get type from *Value element
            self.dtype = interface._schema[self.vtype].dtype

        if self.gain or self.offset:
            self.makeGainOffsetFormat()
        else:
            self._displayFormat = self.makeExpression(self.displayFormat, "displayFormat")
            self._valueFormat = self.makeExpression(self.valueFormat, "valueFormat")

        if not self.label:
            self.label = self._generateLabel(self.configId)

        if not self.tooltip:
            self.tooltip = self.DEFAULT_LABELS.get(self.configId, ('', ''))[1]

        self.configValue = value

        self._fromFile = None  # Indicates value was read from file, set during load
        self._changed = False  # Overrides item value change detection
        self._originalValue = self.value  # Part of change detection


    def parseOptions(self, options: list) -> dict:
        """ Parse a list of `<EnumOption>` elements into a list of value
            choices. Note: also sets `vtype` if not already set by the
            default *Value element.
        """
        parsedoptions = {}
        for n, option in enumerate(options):
            # EnumOption values default to their index in the list
            value = 1 << n if self.element.name.endswith('BitField') else n
            for k, v in option.items():
                if k.endswith('Value'):
                    value = v
                    if self.vtype is None:
                        self.vtype = k
            parsedoptions[value] = option.get('Label', None)
        return parsedoptions


    def makeExpression(self, exp: Optional[str], name: str = ""):
        """ Helper method for compiling an expression in a string into a code
            object that can later be used with `eval()`. Used internally.

            :param exp: The conversion function, as a Python string.
            :param name: The name of the expression (i.e., "displayFormat" or
                "valueFormat"), embedded in the resulting code object. Mostly
                for debugging.
        """
        if exp is None:
            # No expression defined: value is returned unmodified (it matches
            # the config item's type)
            return self.noEffect
        elif exp == '':
            # Expression element exists, but empty: Config item generates no
            # output. Usually for GUI interactivity (disabling one widget
            # based on another, like CheckGroups).
            return None

        # Create a nicely formatted, informative string for the compiled
        # expression's "filename" and for display if the expression is bad.
        idstr = "(ID 0x{:02x}) ".format(self.configId) if self.configId else ""
        msg = "{}{}".format(idstr, name)

        if not isinstance(exp, str):
            # Probably won't occur, but just in case...
            logger.debug("Ignoring bad value for {}: {!r} ({})".format(idstr, exp, type(exp)))
            return

        try:
            return compile(exp, "<{}>".format(msg), "eval")
        except SyntaxError as err:
            logger.error("Ignoring bad expression for {}: {!r}".format(msg, err))
            return self.noEffect


    def makeGainOffsetFormat(self):
        """ Helper method for generating `displayFormat` and `valueFormat`
            expressions using the field's `gain` and `offset`. Used internally.
        """
        gain = 1.0 if self.gain is None else self.gain
        offset = 0.0 if self.offset is None else self.offset

        self.displayFormat = "(x + {:.8f}) * {:.8f}".format(offset, gain)
        self.valueFormat = "x / {:.8f} - {:.8f}".format(gain, offset)

        self._displayFormat = self.makeExpression(self.displayFormat, "displayFormat")
        self._valueFormat = self.makeExpression(self.valueFormat, "valueFormat")


    def __repr__(self):
        msg = "{} ID 0x{:06x}".format(type(self).__name__, self.configId)
        if self.label:
            msg = "{}: {!r}".format(msg, self.label)
        if self.dtype:
            dtype = self.dtype.__name__
            bits = "Bit" in self.element.name
            if self._value is not None:
                val = bin(self._value) if bits else repr(self._value)
                changed = "" if self._value == self._originalValue else "*"
                msg = "{} ({}={}){}".format(msg, dtype, val, changed)
            else:
                msg = "{} ({})".format(msg, dtype)
        return "<{}>".format(msg)


    @property
    def value(self) -> Any:
        """ The configuration item value, in standard engineering units (Pa,
            g, degrees, etc.) where applicable. """
        return self._value


    @value.setter
    def value(self, v: Any):
        """ Set the configuration item value, in engineering units (Pa,
            g, degrees, etc.) where applicable.
        """
        if v is None or (self.interface and not self.interface.validate):
            self._value = v
            return

        if self.element.name.endswith('EnumField') and self.options and v not in self.options:
            raise ValueError("Invalid value for {}, must be one of {}"
                             .format(self, tuple(self.options)))
        elif isinstance(v, str) and len(v) > self.maxLength:
            raise ValueError("Invalid value for {}, max string length is {}"
                             .format(self, self.maxLength))
        elif isinstance(v, (int, float)) and not self.min <= v <= self.max:
            if self.min == float('-inf'):
                msg = "<= {}".format(self.max)
            elif self.max == float('inf'):
                msg = ">= {}".format(self.min)
            else:
                msg = "{} <= v <= {}".format(self.min, self.max)
            raise ValueError("Invalid value for {}, must be {}"
                             .format(self, msg))
        else:
            self._value = v


    @property
    def configValue(self) -> Any:
        """ The item's value, in the config file's native units. """
        # Null string valueFormat
        if self._value is None or self.valueFormat == '':
            return None
        return eval(self._valueFormat, {'x': self._value})


    @configValue.setter
    def configValue(self, v: Any):
        """ Set the item's value using the config file's native units. """
        if v is None:
            self._value = v
        else:
            self._value = eval(self._displayFormat, {'x': v})


    @property
    def default(self) -> Any:
        """ The configuration item's default value. """
        if self._default:
            return eval(self._valueFormat, {'x': self._default})


    @property
    def changed(self) -> bool:
        """ Has the value of the ConfigItem changed? `True` if the current
            value differs from the previous value. This can also be manually
            set to `True` or `False`, although if set `False`, changing the
            ConfigItem's value afterwards will override it.
        """
        return self._changed or self._value != self._originalValue


    @changed.setter
    def changed(self, changed: bool):
        if not changed:
            self._originalValue = self._value
        self._changed = changed


    def revert(self) -> bool:
        """ Reset the item to the value read from the config file, or `None`
            if the item's value was not from a config file.

            :return: `True` if the value was reverted, `False` if the value was
                unchanged.
        """
        if self.configValue != self._fromFile:
            self.configValue = self._fromFile
            self.changed = False
            return True
        return False


    def dump(self) -> Union[None, dict]:
        """ Generate a dictionary containing the item's config ID and value.
            Used when generating a new config file.

            :return: A 2 item dictionary if `value` is not `None`, else
                `None`.
        """
        if self.value is not None:
            return {'ConfigID': self.configId,
                    self.vtype: self.configValue}
        return None


# ===========================================================================
#
# ===========================================================================

class ConfigInterface:
    """
    Base class for mechanisms to access/modify device configuration.

    :ivar config: Device configuration data (e.g., data read from the config
        file). This may be set manually to override defaults and/or existing
        configuration data. Manual setting must be done after instantiation
        and before accessing config elements via the interface's `item`
        attribute.
    :ivar configUi: Device configuration UI information. May be set manually
        to override defaults. Manual setting must be done after instantiation
        and before accessing config elements via `item`.
    :ivar unknownConfig: A dictionary of configuration item types and values,
        keyed by Config ID. Items read from the configuration file that do
        not match items in the ``ConfigUI`` data go into the dictionary.
        These will (by default) be written back to the config file verbatim.
        `unknownConfig` may also be used to manually add arbitrary items to
        the config file.
    """

    def __init__(self, device: "Recorder"):
        """ `ConfigInterface` instances are rarely (if ever) explicitly
            created; the parent `Recorder` object will create the
            appropriate `ConfigInterface` when its `config` property is
            first accessed.

            :param device: The Recorder to configure.
        """
        self._schema = loadSchema('mide_config_ui.xml')
        self.device: Optional["Recorder"] = device
        self.configUi: Optional[MasterElement] = None
        self.config: Optional[MasterElement] = None
        self._items: Dict[int, ConfigItem] = {}
        self._allitems: Dict[int, ConfigItem] = {}  # for debugging, mostly

        # Config values from the loaded configuration data that don't have
        # a corresponding field in the ConfigUI data. Keyed by ConfigID,
        # values are tuples of (*Value element name, value). This can be
        # modified directly to add new/custom configuration values.
        self.unknownConfig: Dict[int, Tuple[str, Any]] = {}

        # For future use
        self.postConfigMsg = None

        # For disabling validation of item values. Setting to `False` will
        # ignore the item's min/max and options. Primarily for testing.
        self.validate = True

        # The format version of the last config data read.
        self.configVersionRead = None
        self._supportedConfigVersions = None


    def close(self) -> bool:
        """
        Close the interface. Only applicable to subclasses with a persistent
        connection. Fails silently.

        :return: `True` if the connection was reset (or the interface type
            has no persistent connection).
        """
        return True


    @property
    def supportedConfigVersions(self):
        """ A tuple of configuration data format versions supported by
            the interface. 1 is the original SlamStick version, only
            supported on old hardware/firmware. The current version is 2.
        """
        return (2,)


    @property
    def items(self) -> Dict[int, ConfigItem]:
        """ All defined configuration items for the device, keyed by
            Config ID.
        """
        if not self.configUi:
            self._items.clear()
            self._allitems.clear()
            self.configUi = self.getConfigUI()
            self.parseConfigUI(self.configUi)

        if self.config is None:
            try:
                self.loadConfig()
            except Exception as err:
                warnings.warn('Error when reading config file: {!r}'.format(err))

        return self._items


    @items.setter
    def items(self, items: Dict[int, ConfigItem]):
        self._items = items
        self._allitems.update(items)


    @classmethod
    def hasInterface(cls, device: "Recorder") -> bool:
        """ Determine if a device supports this `ConfigInterface` type.

            :param device: The Recorder to check.
            :return: `True` if the device supports the interface.
        """
        if device.isRemote:
            return False
        if getattr(device, "_config", None) is not None:
            return True
        return ui_defaults.getDefaultConfigUI(device) is not None


    @property
    def available(self) -> bool:
        """ Is the device currently ready for configuration?

            Note: This is intended for future configuration systems. Since
            configuration is currently applied via the filesystem, it is
            functionally the same as `Recorder.available`.
        """
        # For now, this is basically the same as the device availability,
        # but allows for future config interfaces (e.g., remote and/or
        # wireless, etc.)
        return self.device.available


    def parseConfigUI(self,
                      configUi: Union[Document, Element]):
        """ Recursively process CONFIG.UI data to populate the interface's
            dictionaries of configuration items.

            :param configUi: A parsed `CONFIG.UI` EBML document or element.
        """
        for el in configUi:
            if el.name == 'PostConfigMessage':
                self.postConfigMsg = el.value
                continue

            if not isinstance(el, MasterElement):
                continue

            # Note: all MasterElement subclasses have the attribute `name`
            if el.name.endswith("Field") or el.name == 'CheckGroup':
                data = el.dump()
                if 'ConfigID' in data:
                    item = ConfigItem(self, el, data)
                    self._allitems[item.configId] = item

                    # Exclude items that don't generate config file values
                    if item.valueFormat != '':
                        self._items[item.configId] = item

            if el.name in ('ConfigUI', 'Tab') or 'Group' in el.name:
                self.parseConfigUI(el)


    def getChanges(self) -> List[ConfigItem]:
        """ Has the configuration data been modified? """
        # Deliberately check all `changed` (it resets when gotten)
        return [item for item in self.items.values() if item.changed]


    def revert(self):
        """ Reset all configuration values to what was originally read from
            the configuration file, or `None` if the value was never read
            from the file.
        """
        for item in self.items.values():
            if item.revert():
                logger.debug('Reverted {!r}'.format(item))


    def _parseConfig(self,
                     origConfig: dict,
                     default: Optional[dict] = None) -> Union[Dict[int, Any], None]:
        """ Helper method to parse a dictionary of dumped EBML
            ``RecorderConfigurationList`` data into a simple dictionary of
            values keyed by ConfigID. Used internally.

            :param origConfig: The unprocessed dictionary dumped from a
                configuration EBML file.
            :param default: A dictionary of default config values.
        """
        config = {} if default is None else default.copy()

        root = origConfig.get('RecorderConfigurationList', {})
        if not root:
            return None

        for item in root.get('RecorderConfigurationItem', []):
            configId = item.get('ConfigID', None)
            if not configId:
                continue

            for k, v in item.items():
                if k.endswith('Value') and v is not None:
                    config[configId] = bool(v) if "Boolean" in k else v
                    break

        return config


    def _makeConfig(self,
                    unknown: bool = True,
                    version: Optional[int] = None) -> Dict[str, Any]:
        """ Generate a dictionary of configuration data, suitable for EBML
            encoding.

            Note: this is currently used directly by another project (the
            config GUI's exporter). Be careful modifying until import/export
            has been moved to this package and the config GUI is updated.

            :param unknown: If `True`, include configuration items in the
                `ConfigInterface`'s `unknownConfig`; items read from the
                configuration file but have IDs that do not correspond to
                fields in the device's ``ConfigUI`` data.
            :param version: The version of configuration data to use.
            :return: A dictionary of configuration values, ready for encoding
                as EBML.
        """
        supportedVersions = self.supportedConfigVersions
        if version is not None and version not in supportedVersions:
            raise ConfigError("Device does not support config data version {}, only {}".
                              format(version, supportedVersions))

        config = [item.dump() for item in self.items.values()
                  if item.dump() is not None]
        if unknown:
            for k, v in self.unknownConfig.items():
                config.append({'ConfigID': k, v[0]: v[1]})

        return {'RecorderConfigurationList':
                    {'RecorderConfigurationItem': config}}


    def getConfigUI(self) -> Union[Document, MasterElement]:
        """ Get the device's ``ConfigUI`` data.
        """
        raise NotImplementedError("getConfigUI() not implemented")


    def getConfig(self) -> Union[None, Document, MasterElement]:
        """ Low-level method that retrieves the device's config EBML (e.g.,
            the contents of a real device's `config.cfg` file), if any.
        """
        raise NotImplementedError("getConfig() not implemented")


    def getConfigValues(self,
                        original: bool = False,
                        defaults: bool = False,
                        none: bool = False,
                        unknown: bool = True) -> Dict[int, Any]:
        """ Get the device configuration as a simple dictionary of values
            keyed by config ID.

            :param original: If `True`, return only the values read from the
                device configuration. Overrides the other parameters.
            :param defaults: If `True`, include items with their default
                values, not only those explicitly set or read from a file.
            :param none: If `False`, exclude items with values of `None`.
            :param unknown: If `True`, include values read from the
                config file that do not correspond to know configuration
                items.
        """
        # TODO: this seems really kludgey. Maybe split into different methods?
        if original:
            if not self.config:
                return {}

            return self._parseConfig(self.config.dump())

        conf = {item.configId: item.value for item in self.items.values()
                if ((defaults or item.configValue != item._default) and
                    (none or item.value is not None))}

        if unknown:
            for k, v in self.unknownConfig.items():
                conf[k] = v[1]

        return conf


    def applyConfig(self,
                    unknown: bool = True,
                    version: Optional[int] = None):
        """ Apply (save) configuration data to the device.

            :param unknown: If `True`, include values that do not correspond
                to known configuration items (e.g., originally read from the
                config file).
            :param version: The version of configuration data to use, if the
                device supports more than one. Defaults to the latest
                version supported.
        """
        raise NotImplementedError("applyConfig() not implemented")


    def saveAs(self,
               filename: Union[str, Path],
               unknown: bool = True,
               version: Optional[int] = None):
        """ Save the raw configuration data to a file (e.g., ``config.cfg``).
            The file saved is the equivalent of the recorder's native config
            data, and can only be applied to the original recorder (or one of
            the same model). To export configuration data for use on another
            device, use `endaq.device.configio.exportConfig()`.

            :param filename: The name of the output file. By convention, the
                extension is ``.cfg``, but this is not enforced.
            :param unknown: If `True`, include values that do not correspond
                to known configuration items (e.g., originally read from the
                config file).
            :param version: The version of configuration data to use, if the
                device supports more than one. Defaults to the latest
                version supported.
        """
        data = self._makeConfig(unknown=unknown, version=version)
        with open(filename, 'wb') as f:
            loadSchema('mide_ide.xml').encode(f, data)


    def loadConfig(self, config: Optional[MasterElement] = None):
        """ Process a device's configuration data.

            :param config: Optional, explicit configuration EBML data to
                process. If none is provided, the data retrieved by
                `getConfig()` will be used.
        """
        if config is None:
            config = self.config or self.getConfig()

        if not config:
            self.config = {}
            return

        dump = config.dump()

        root = dump.get('RecorderConfigurationList', dump)
        if root:
            for item in root.get('RecorderConfigurationItem', []):
                k = item['ConfigID']
                v = next(filter(lambda x: 'Value' in x[0], item.items()))
                if k in self._items:
                    self._items[k].configValue = v[1]
                    self._items[k].changed = False
                    self._items[k]._fromFile = v[1]
                else:
                    self.unknownConfig[k] = v

        self.config = config
        self.configVersionRead = 2  # Future: detect version?


    # =======================================================================
    #
    # =======================================================================

    def _getitem(self, item: Union[int, str]) -> ConfigItem:
        """ Get a configuration item, specified either by config ID or by label
            (if it has one).

            Note that labels are not guaranteed to be unique, but config IDs
            are. Strings must exactly match label text. Using config IDs is
            therefore recommended.

            :param item: The config ID or label of a configuration item.
            :return: The indicated `ConfigItem`.
        """
        if item in self.items:
            return self.items[item]

        s = hex(item) if isinstance(item, int) else repr(item)
        raise KeyError(item, "Config item {} not in CONFIG.UI data"
                       .format(s))


    def _setitem(self, item: Union[int, str], value: Optional[Any]):
        """ Set a configuration item, specified either by config ID or by label
            (if it has one).

            Note that labels are not guaranteed to be unique, but config IDs
            are. Strings must exactly match label text. Using config IDs is
            therefore recommended.

            :param item: The config ID or label of a configuration item.
        """
        # Somewhat redundant w/ VirtualConfigInterface, but just to be safe
        if self.device and self.device.isVirtual:
            raise ConfigError('Cannot configure virtual devices')
        self._getitem(item).value = value


    # =======================================================================
    # Simple, standard config items, giving names to config IDs. A couple do
    # a small amount of modification for convenience (e.g., UTC Offset).
    # =======================================================================

    @property
    def name(self) -> Union[str, None]:
        """ User-entered name of the device. """
        return self._getitem(0x08ff7f).value

    @name.setter
    def name(self, n: Optional[str]):
        self._setitem(0x08ff7f, n)

    @property
    def notes(self) -> Union[str, None]:
        """ User-entered notes/description of the device. """
        return self._getitem(0x09ff7f).value

    @notes.setter
    def notes(self, n: Optional[str]):
        self._setitem(0x09ff7f, n)

    @property
    def tags(self) -> Union[str, None]:
        """ The device's recording tags (comma-separated string), stored in
            each recording. Primarily used for organization on enDAQ Cloud.
        """
        return self._getitem(0x17ff7f).value

    @tags.setter
    def tags(self, t):
        self._setitem(0x17ff7f, t)

    @property
    def pluginAction(self) -> Union[int, None]:
        """ The ID of the device's Plug-In Action (what happens when
            attached by USB).
        """
        return self._getitem(0x0aff7f).value

    @pluginAction.setter
    def pluginAction(self, action: int):
        self._setitem(0x0aff7f, action)

    @property
    def pluginActions(self) -> Dict[int, str]:
        """ The IDs and descriptions of all known Plug-In Action options.
            Read only.
        """
        return self._getitem(0x0aff7f).options

    @property
    def buttonMode(self) -> Union[int, None]:
        """ The ID of the device's 'button mode' (what happens when the
            device's primary button is pressed).
        """
        return self._getitem(0x10ff7f).value

    @buttonMode.setter
    def buttonMode(self, mode: int):
        self._setitem(0x10ff7f, mode)

    @property
    def buttonModes(self) -> Dict[int, str]:
        """ The IDs and descriptions of all known Button Mode options.
            Read only.
        """
        return self._getitem(0x10ff7f).options

    @property
    def utcOffset(self) -> Union[float, int, None]:
        """ The recorder's local offset from UTC, in minutes. """
        # Convert from seconds to hours
        return self._getitem(0x0bff7f).value / 3600

    @utcOffset.setter
    def utcOffset(self, offset: Optional[float]):
        # Convert from hours to seconds
        self._setitem(0x0bff7f, int(offset * 3600))

    @property
    def recordingDir(self) -> Union[str, None]:
        """ The name of the directory (on the device) where recordings are saved. """
        return self._getitem(0x14ff7f).value

    @recordingDir.setter
    def recordingDir(self, dirname: Optional[str]):
        self._setitem(0x14ff7f, dirname)

    @property
    def recordingPrefix(self) -> Union[str, None]:
        """ Prefix string for all recording filenames. """
        return self._getitem(0x15ff7f).value

    @recordingPrefix.setter
    def recordingPrefix(self, prefix: Optional[str]):
        self._setitem(0x15ff7f, prefix)

    @property
    def recordingTimeLimit(self) -> Union[int, None]:
        return self._getitem(0x0dff7f).value

    @recordingTimeLimit.setter
    def recordingTimeLimit(self, t: Optional[int]):
        self._setitem(0x0dff7f, t)

    @property
    def recordingSizeLimit(self) -> Union[int, None]:
        return self._getitem(0x11ff7f).value

    @recordingSizeLimit.setter
    def recordingSizeLimit(self, t: Optional[int]):
        self._setitem(0x11ff7f, t)


    # =======================================================================
    # More complex (but standard) configuration items
    # =======================================================================

    def _encodeChannel(self,
                       channel: Union[Channel, SubChannel]) -> int:
        """ Create the low 2 bytes of a channel-specific config ID (low byte
            is channel ID, high byte is subchannel ID or 0xFF if not a
            `SubChannel`).

            :param channel: The channel or subchannel to encode
            :return: The encoded channel ID
        """
        if isinstance(channel, SubChannel):
            return channel.id << 8 | channel.parent.id
        elif isinstance(channel, Channel):
            return channel.id | 0xff00
        else:
            raise TypeError("Expected a Channel or SubChannel object, "
                            "got {}".format(type(channel)))


    def enableChannel(self,
                      channel: Union[Channel, SubChannel],
                      enabled: bool = True):
        """ Enable or disable a `Channel` or `SubChannel`.

            Due to the way in which they operate, some sensor's `SubChannels`
            cannot be individually configured, and changes must be made to
            the parent `Channel`. The opposite is true for other sensors -
            primarily the analog ones - which can only be configured at the
            `SubChannel` level. Attempting to configure at the wrong 'level'
            will raise a ``ConfigError``.

            :param channel: The channel or subchannel to enable/disable.
            :param enabled: `True` to enable the channel/subchannel recording.
        """
        configId = self._getChannelConfigId(0x010000, channel)
        enItem = self._getitem(configId)

        # Some subchannels (analog) are enabled explicitly with their own
        # config ID. Others (mostly digital) use bits in a single config item.
        if enItem.element.name.endswith('BitField'):
            # Some Channels without individually-configurable SubChannels use
            # a one-item BitField.
            if len(enItem.options) == 1:
                if isinstance(channel, SubChannel):
                    raise ConfigError('ConfigUI {} not applicable to {}; '
                                      'use a Channel instead'
                                      .format(enItem, channel))
                enabled = int(enabled)

            else:
                if not isinstance(channel, SubChannel):
                    raise ConfigError('ConfigUI {} not applicable to {}; '
                                      'use a SubChannel instead'
                                      .format(enItem, channel))

                val = enItem.value
                if val is None:
                    val = 2 ** len(channel.parent.children) - 1
                bit = (1 << channel.id)
                if enabled:
                    # Set
                    enabled = val | bit
                elif val & bit:
                    # Clear if set
                    enabled = val ^ bit
                else:
                    # No change
                    enabled = val

        enItem.value = enabled


    def isEnabled(self,
                  channel: Union[Channel, SubChannel]) -> bool:
        """ Is the `Channel` or `SubChannel` enabled?

            Due to the way in which they operate, some sensor's `SubChannels`
            cannot be individually configured, and changes must be made to
            the parent `Channel`. The opposite is true for other sensors -
            primarily the analog ones - which can only be configured at the
            `SubChannel` level. Attempting to configure at the wrong 'level'
            will raise a ``ConfigError``.

            :param channel: The `Channel` or `SubChannel` to check.
            :return: `True` if configured to record.
        """
        configId = self._getChannelConfigId(0x010000, channel)
        enItem = self._getitem(configId)
        en = enItem.value
        if enItem.element.name.endswith('BitField') and isinstance(channel, SubChannel):
            en = True if en is None else (int(en) & (1 << channel.id))
        return bool(en)


    def _getChannelConfigId(self,
                            configType: int,
                            channel: Union[Channel, SubChannel]) -> int:
        """ Get the config ID of a specific type for a channel. If the config
            ID for the specific subchannel doesn't exist, the config ID for
            the channel is returned (although it might not exist, either).
            For things like BitField values, in which there is one config
            item for the channel, and the bits of the value represent the
            subchannels.

            :param configType: A 24b value, with the high byte representing
                the configuration item type. Bits 0-15 are 0.
            :param channel: The channel or subchannel to configure.
            :return: The channel/subchannel configuration ID.
        """
        configId = configType | self._encodeChannel(channel)
        if configId not in self.items:
            configId |= 0x00ff00
        return configId


    def _getChannel(self, configId: int) -> Union[Channel, SubChannel, None]:
        """ Get the Channel/SubChannel corresponding to a configuration ID.
        """
        ch = configId & 0xFF
        subCh = configId >> 8 & 0xFF

        channel = self.device.channels.get(ch, None)
        if not channel:
            logger.debug('No channel for ConfigID {:x}'.format(configId))
            return None

        if subCh == 0xFF:
            return channel

        if subCh > len(channel.children):
            logger.debug('No subchannel for ConfigID {:x}'.format(configId))
            return None

        return channel[subCh]


    def setTrigger(self,
                   channel: Union[Channel, SubChannel],
                   **kwargs):
        """ Set the trigger for a `Channel` or `Subchannel`.

            :param channel: The channel or subchannel to configure.
            :keyword low: The trigger's low threshold value (if applicable).
            :keyword high: The trigger's high threshold value (if applicable).
                Also used for sensors that have only an absolute threshold
                trigger (e.g., a shock trigger).
            :keyword enabled: `True` to enable the channel/subchannel trigger.
                Threshold arguments are Values of `high` and `low` are optional if `False`.
        """
        loId = self._getChannelConfigId(0x030000, channel)
        hiId = self._getChannelConfigId(0x040000, channel)
        enId = self._getChannelConfigId(0x050000, channel)

        # Some sensors have a main channel-level trigger enable as well as
        # individual subchannel triggers. These main enables have a different
        # 'prefix' byte (0x07).
        if enId not in self.items and enId & 0x00ff00:
            enId = self._getChannelConfigId(0x070000, channel)

        # Fail before setting if any ConfigID is unknown is bad.
        if 'low' in kwargs and loId not in self.items:
            raise ConfigError('Cannot configure low trigger for {!r}'
                              .format(channel))
        if 'high' in kwargs and hiId not in self.items:
            raise ConfigError('Cannot configure high trigger for {!r}'
                              .format(channel))
        if 'enabled' in kwargs and enId not in self.items:
            raise ConfigError('Cannot configure enable for {!r}'
                              .format(channel))

        if 'low' in kwargs:
            self._setitem(loId, kwargs['low'])
        if 'high' in kwargs:
            self._setitem(hiId, kwargs['high'])
        if 'enabled' in kwargs:
            enItem = self._getitem(enId)
            if enItem.element.name.endswith('BitField') and isinstance(channel, SubChannel):
                en = enItem.value or 0
                bit = 1 << channel.id
                en = en | bit if kwargs['enabled'] else en ^ bit
            else:
                en = int(kwargs['enabled'])

            self._setitem(enId, en)


    def getTrigger(self,
                   channel: Union[Channel, SubChannel]) -> dict:
        """ Get the trigger information for a `Channel` or `SubChannel`.

            :param channel: The `Channel` or `SubChannel` to get.
            :return: A `dict` of trigger parameters, usable as keyword
                arguments for `setTrigger()`.
        """
        loId = self._getChannelConfigId(0x030000, channel)
        hiId = self._getChannelConfigId(0x040000, channel)
        enId = self._getChannelConfigId(0x050000, channel)

        trig = {}
        if loId in self.items:
            trig['low'] = self._getitem(loId).value
        if hiId in self.items:
            trig['high'] = self._getitem(hiId).value

        enItem = self._getitem(enId)
        en = enItem.value
        if enItem.element.name.endswith('BitField') and isinstance(channel, SubChannel):
            # Enabled is usually the default; change per-case if/when not
            en = True if en is None else bool(en & (1 << channel.id))
        trig['enabled'] = en

        return trig


    def setSampleRate(self,
                      channel: Channel,
                      sampleRate: float):
        """ Set the sample rate of a `Channel`.

            :param channel: The `Channel` to set.
            :param sampleRate: The new sampling rate, in hertz.
        """
        configId = 0x020000 | self._encodeChannel(channel)
        self._getitem(configId).value = sampleRate


    def getSampleRate(self,
                      channel: Channel) -> float:
        """ Get the sample rate of a `Channel`.

            :param channel: The `Channel` to get.
            :return: The sampling rate, in hertz.
        """
        configId = self._encodeChannel(channel) | 0x020000
        item = self._getitem(configId)
        return item.value if item.value is not None else item.default


    def getTriggers(self) -> List[ConfigItem]:
        """ Get all settable triggers (threshold, high, and/or low).

            :return: A list of all trigger configuration items.
        """
        return [v for v in self.items.values()
                if v.configId & 0xff0000 in (0x030000, 0x040000)]


# ===========================================================================
#
# ===========================================================================

class VirtualConfigInterface(ConfigInterface):
    """
    Read-only configuration interface for 'virtual' recorders (i.e., ones
    instantiated from IDE file metadata).
    """

    @classmethod
    def hasInterface(cls, device: "Recorder") -> bool:
        """
        Determine if a device supports this `ConfigInterface` type.

        :param device: The Recorder to check.
        :return: `True` if the device supports the interface.
        """
        return device.isVirtual and super().hasInterface(device)


    @property
    def available(self) -> bool:
        """ Is the device currently ready for configuration?
        """
        # Never!
        return False


    def getConfigUI(self) -> Union[Document, MasterElement]:
        """ Load the virtual device's ``ConfigUI`` data.
        """
        # Use existing data, or data taken from source file
        self.configUi = self.configUi or getattr(self.device, '_configUi', None)
        if self.configUi:
            return self.configUi
        else:
            ebml = ui_defaults.getDefaultConfigUI(self.device)
            self.configUi = self._schema.loads(ebml)

        if not self.configUi:
            raise IOError(errno.ENOENT, "No default ConfigUI found for {}"
                          .format(self.device.partNumber))

        return self.configUi


    def getConfig(self) -> Union[None, Document, MasterElement]:
        """ Low-level method that retrieves the device's config EBML (e.g.,
            the contents of a real device's ``config.cfg`` file), if any.
        """
        # This will have been cached when Recorder.fromRecording() was called
        return self.device._configData


    def applyConfig(self, **kwargs):
        """ Apply modified configuration data to the device. Not supported on
            virtual devices!
        """
        raise UnsupportedFeature("Virtual devices cannot be configured")


    def _setitem(self, item: Union[int, str], value):
        """ Set a configuration item, specified either by config ID or by
            label (if it has one). Not supported on virtual devices!

            Note that labels are not guaranteed to be unique, but config IDs
            are. Strings must exactly match label text. Using config IDs is
            therefore recommended.

            :param item: The config ID or label of a configuration item.
        """
        raise UnsupportedFeature("Virtual devices cannot be configured")


    def revert(self):
        """ Reset all configuration values to what was originally read from
            the configuration file. Not supported on virtual devices!
        """
        raise UnsupportedFeature("Virtual devices cannot be configured")


# ===========================================================================
#
# ===========================================================================

class FileConfigInterface(ConfigInterface):
    """
    Standard configuration interface: uses the device's ``CONFIG.UI`` file
    (if present), and reads/writes a local device's ``config.cfg`` file.

    This interface also handles converting to and from the legacy
    configuration file format (version 1).
    """

    @classmethod
    def hasInterface(cls, device: "Recorder") -> bool:
        """
        Determine if a device supports this `ConfigInterface` type.

        :param device: The Recorder to check.
        :return: `True` if the device supports the interface.
        """
        if device.isVirtual or device.isRemote:
            return False

        # Very simple initial check: is there a CONFIG.UI file?
        # Unlikely to fail, but in a `try` just in case.
        try:
            if os.path.isfile(device.configUIFile):
                return True
        except IOError:
            pass

        return super().hasInterface(device)


    @property
    def supportedConfigVersions(self):
        """ A tuple of configuration file format versions supported by
            the interface.
        """
        if self._supportedConfigVersions is not None:
            return self._supportedConfigVersions

        mcu = self.device.getInfo('McuType', 'EFM32GG330')
        if not mcu.startswith("EFM32GG330"):
            return ConfigInterface.supportedConfigVersions.fget(self)

        vers = (1,)

        if self.device.firmwareVersion > 14:
            vers += ConfigInterface.supportedConfigVersions.fget(self)

        self._supportedConfigVersions = vers
        return vers


    def _makeConfig(self,
                    unknown: bool = True,
                    version: Optional[int] = None,
                    defaults: bool = False) -> Dict[str, Any]:
        """ Generate a dictionary of configuration data, suitable for EBML
            encoding.

            Note: this is currently used directly by another project (the
            config GUI's exporter). Be careful modifying until import/export
            has been moved to this package and the config GUI is updated.

            :param unknown: If `True`, include configuration items in the
                `ConfigInterface`'s `unknownConfig`; items read from the
                configuration file but have IDs that do not correspond to
                fields in the device's ``ConfigUI`` data.
            :param version: The version of configuration data to use.
            :param defaults: If `True`, include config values that have not
                been explicitly set (i.e. still their default value).
            :return: A dictionary of configuration values, ready for encoding
                as EBML.
        """
        supportedVersions = self.supportedConfigVersions
        if version is not None and version not in supportedVersions:
            raise ConfigError("Device does not support config data version {}, only {}".
                              format(version, supportedVersions))

        version = version or supportedVersions[-1]

        if version != 1:
            return super()._makeConfig(unknown=unknown)

        logger.debug('Writing legacy config file for {!r}'.format(self.device))
        vals = self.getConfigValues(defaults=defaults, original=False, unknown=unknown)
        return legacy.generateLegacyConfig(vals, self.device)


    def getConfigUI(self) -> Union[Document, MasterElement]:
        """ Load the device's ``ConfigUI`` data.
        """
        if not self.configUi:
            if os.path.isfile(self.device.configUIFile):
                with open(self.device.configUIFile, 'rb') as f:
                    self.configUi = self._schema.loads(f.read())
            else:
                ebml = ui_defaults.getDefaultConfigUI(self.device)
                if not ebml:
                    raise IOError(errno.ENOENT, "No default ConfigUI found for {}"
                                  .format(self.device))
                self.configUi = self._schema.loads(ebml)

        return self.configUi


    def getConfig(self) -> Union[None, Document, MasterElement]:
        """ Low-level method that retrieves the device's config EBML (e.g.,
            the contents of a real device's `config.cfg` file), if any.
        """
        if not os.path.isfile(self.device.configFile):
            return None

        try:
            with open(self.device.configFile, 'rb') as f:
                data = f.read()
            if data:
                return loadSchema('mide_ide.xml').loads(data)

            logger.debug('No config data could be read (device not configured?), ignoring')

        except (DeviceError, IOError, NotImplementedError) as err:
            warnings.warn("{}.getConfig(): ignoring possibly expected exception {!r}"
                          .format(type(self).__name__, err))

        return None


    def loadConfig(self, config: Optional[MasterElement] = None):
        """ Process a device's configuration data.

            :param config: Optional, explicit configuration EBML data to
                process. If none is provided, the data retrieved by
                `getConfig()` will be used.
        """
        config = config or self.config or self.getConfig()
        if not config:
            self.config = {}
            return

        if config[0].name == "RecorderConfiguration":
            logger.debug("Reading legacy config file data from {}"
                         .format(self.device.configFile))
            versionRead = 1
            self._originalConfig = config  # for testing
            config = legacy.convertConfigFile(config, self)
        else:
            versionRead = 2

        super().loadConfig(config)
        self.configVersionRead = versionRead


    def applyConfig(self,
                    unknown: bool = True,
                    clear: bool = True,
                    version: Optional[int] = None):
        """ Save modified configuration data to the device.

            :param unknown: If `True`, include values read from the config
                file that did not correspond to known configuration items.
            :param clear: If `True`, mark all items as unchanged after
                application.
            :param version: The version of configuration data to use, if the
                device supports more than one. Defaults to the latest
                version supported.
        """
        if not self.available:
            raise IOError(errno.ENOENT, "Could not find {}; is it connected?"
                          .format(self.device))

        # Do encoding before opening the file, so it can fail safely and not
        # affect any existing config file.
        config = self._makeConfig(unknown, version=version)
        configEbml = loadSchema('mide_ide.xml').encodes(config, headers=False)

        try:
            util.makeBackup(self.device.configFile)
            with open(self.device.configFile, 'wb') as f:
                f.write(configEbml)

            if clear:
                for item in self.items.values():
                    item.changed = False

        except Exception:
            # Write failed, restore old config file
            util.restoreBackup(self.device.configFile, remove=False)
            raise


# ===========================================================================
#
# ===========================================================================

#: A list of all `ConfigInterface` types, used when finding a device's
#   interface. `VirtualConfigInterface` should go last. New interface types
#   defined elsewhere should append/insert themselves into this list (before
#   their superclass, if their `hasInterface()` is more specific).
INTERFACES = [FileConfigInterface, VirtualConfigInterface]



"""
Functions for filtering data by measurement types, and singleton objects
representing different measurement types.
"""

__all__ = ['ANY', 'ACCELERATION', 'ALTITUDE', 'ANG_RATE', 'AUDIO', 'DIRECTION',
           'FREQUENCY', 'GENERIC', 'GYRO', 'HUMIDITY', 'LIGHT', 'LOCATION',
           'MAGNETIC', 'ORIENTATION', 'PRESSURE', 'ROTATION', 'SPEED',
           'TEMPERATURE', 'TIME', 'VOLTAGE',
           'get_measurement_type', 'filter_channels']

from fnmatch import fnmatch
from shlex import shlex
from typing import List, Dict, Tuple, Union

from idelib.dataset import Channel, SubChannel

# ============================================================================
#
# ============================================================================


class MeasurementType:
    """ Singleton marker object for filtering channels by measurement type.
    """
    # TODO: Fix nomenclature. "Singleton" may not be the correct term; there are multiple instances
    #   of MeasurementType, but not multiple *duplicate* ones.

    types = {}
    verbose = False

    def __new__(cls, name, key, labels=None, doc=None):
        """ Create and return a new object, if one with a matching `key` does
            not already exist. If it does, the previous instance is returned.
            Called prior to `__init__()`.

            :param name: The display name of the measurement type.
            :param key: The abbreviated 'key' string for the type. Functions
                that use `MeasurementType` objects can also use those key
                strings.
            :param labels: A list/tuple of substrings to match against
                `Channel` and `SubChannel` unit labels, used to identify
                the appropriate `NeasurementType` instance. The `name` is
                automatically included.
            :param doc: A docstring for the `MeasurementType` instance.
            :returns: `MeasurementType`
        """
        key = str(key).lower()
        if key not in cls.types:
            obj = super().__new__(cls)
            obj._name = name
            obj._key = key
            if labels is not None:
                labels = (labels.lower(),) if isinstance(labels, str) else labels
                labels = tuple(labels) if labels else ()
                if name not in labels:
                    labels += (name.lower(),)
            obj._labels = labels
            obj.__doc__ = doc or name
            cls.types[key] = obj
        return cls.types[key]

    def __str__(self):
        return self._key

    def __repr__(self):
        if self.verbose:
            return "<Measurement Type: {self._name} {self.key!r})>".format(self=self)
        return "<Measurement Type: {self._name}>".format(self=self)

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        # For interoperability with 'key' strings (as dict keys, etc.)
        # e.g., `mtype == "acc"`
        return hash(self._key)

    def __add__(self, other):
        # Concatenates as strings, so query-like sequences can be built.
        return "{} {}".format(self, other)

    def __radd__(self, other):
        # Concatenates as strings, so query-like sequences can be built.
        return "{} {}".format(other, self)

    def __sub__(self, other):
        # Concatenates as strings (with this one's negated)
        return "{} -{}".format(self, other)

    def __rsub__(self, other):
        # Concatenates as strings (with this one's negated)
        return "{} -{}".format(other, self)

    def __or__(self, other):
        # Same as __add__(), but convenient for those used to using bitwise OR to combine flags
        return self + other

    def __neg__(self):
        # Negates (appends a `-`) so query-like strings can be built.
        return "-{}".format(self)

    @property
    def name(self):
        return self._name

    @property
    def key(self):
        return self._key

    def match(self, channel):
        """ Does the given object match this measurement type?

            :param channel: The object to test. Can be an `idelib.dataset.Channel`,
                `idelib.dataset.SubChannel`, or a string of a measurement
                type name.
            :return: `True` if the given object uses this measurement type.
                In the case of `idelib.dataset.Channel`, `True` is returned
                if any of its subchannels match.
        """
        if getattr(channel, '_measurementType', channel) == self:
            return True

        if isinstance(channel, Channel):
            if not isinstance(channel, SubChannel):
                return any(self.match(c) for c in channel.children)
            else:
                mt = channel.units[0]
        elif isinstance(channel, str):
            mt = channel
        else:
            raise TypeError("Cannot compare measurement types with {} ({})".format(channel, type(channel)))

        mt = mt.lower()
        return any(label in mt or fnmatch(mt, label) for label in self._labels)

# ============================================================================
#
# ============================================================================


ANY = MeasurementType("Any/all", "*",
    doc="Marker object for matching any/all measurement types",
    labels=("*",))

ACCELERATION = MeasurementType("Acceleration", "acc",
    doc="Marker object for filtering channels with acceleration data",
    labels=())
ORIENTATION = MeasurementType("Orientation", "imu",
    doc="Marker object for filtering channels with rotation/orientation data",
    labels=("quaternion", "euler", "orientation"))
ROTATION = MeasurementType("Rotation", "rot",
    doc="Marker object for filtering channels with angular change rate data",
    labels=("rotation", "gyro"))
AUDIO = MeasurementType("Audio", "mic",
    doc="Marker object for filtering channels with sound level data",
    labels=("mic",))
LIGHT = MeasurementType("Light", "lux",
    doc="Marker object for filtering channels with light intensity data",
    labels=("lux", "uv"))
PRESSURE = MeasurementType("Pressure", "pre",
    doc="Marker object for filtering channels with air pressure data",
    labels=())  # pressures
TEMPERATURE = MeasurementType("Temperature", "tmp",
    doc="Marker object for filtering channels with temperature data",
    labels=())  # temperature
HUMIDITY = MeasurementType("Humidity", "hum",
    doc="Marker object for filtering channels with (relative) humidity data",
    labels=())  # Humidity
LOCATION = MeasurementType("Location", "gps",
    doc="Marker object for filtering channels with location data",
    labels=("pos",))  # GPS
SPEED = MeasurementType("Speed", "spd",
    doc="Marker object for filtering channels with rate-of-speed data",
    labels=("velocity",))  # GPS Ground Speed
TIME = MeasurementType("Time", "epo",
    doc="Marker object for filtering channels with time data",
    labels=("epoch",))  # GPS Epoch Time

# For potential future use
GENERIC = MeasurementType("Generic/Unspecified", "raw",
    labels=("adc", "raw"))
ALTITUDE = MeasurementType("Altitude", "alt",
    doc="Marker object for filtering channels with altitude data",
    labels=())
VOLTAGE = MeasurementType("Voltage", "vol",
    doc="Marker object for filtering channels with voltmeter data",
    labels=("volt",))
DIRECTION = MeasurementType("Direction", "dir",
    doc="Marker object for filtering channels with 2D directional data",
    labels=("compass", "heading"))
MAGNETIC = MeasurementType("Magnetic Field", "emf",
    doc="Marker object for filtering channels with magnetic field strength data",
    labels=("emf", "magnetic"))
FREQUENCY = MeasurementType("Frequency", "fre",
    doc="Marker object for filtering channels with frequency data",
    labels=("rate",))

# Synonyms, for convenience.
# TODO: Include abbreviations (e.g., TEMP = TEMPERATURE) for convenience? Current names are long.
ADC = GENERIC
ANG_RATE = ROTATION
GYRO = ROTATION
RAW = GENERIC
QUATERNION = ORIENTATION


# ============================================================================
#
# ============================================================================

def get_measurement_type(channel: Channel) -> Union[MeasurementType, List[MeasurementType], None]:
    """ Get the appropriate `MeasurementType` object for a given `SubChannel`.
        Calling with a `Channel` returns a list of `MeasurementType` objects,
        with one for each child `SubChannel`.

        :param channel: A `Channel` or `SubChannel` instance (e.g., from a
            `Dataset`).
        :returns: A `MeasurementType` object (for a `SubChannel`), or a list
            of `MeasurentType` objects (one for each child) if a `Channel`
            was supplied.
    """
    mt = getattr(channel, "_measurementType", None)
    if mt:
        return mt

    if channel.children:
        # Note: this caches the MeasurementType in the Channel/SubChannel; the
        # attribute `_measurementType` must be settable
        channel._measurementType = [get_measurement_type(c) for c in channel.children]
        return channel._measurementType
    else:
        for m in MeasurementType.types.values():
            if m == ANY:
                continue
            if m.match(channel):
                channel._measurementType = m
                return m

    return None


def split_types(query: Union[MeasurementType, str]) -> Tuple[set, set]:
    """ Split a string of multiple `MeasurementType` keys (e.g., one generated
        by using addition or subtraction of `MeasurementType` objects and/or
        strings).

        :param query: A `MeasurementType` or a string containing multiple
            `MeasurementType` keys. A key can be excluded by prefixing it with
            a ``-``.
        :returns: A pair of sets of `MeasurementType` instances: ones to
            include, and ones to exclude.
    """
    query = str(query).lower().strip()
    if query == "*":
        return set(MeasurementType.types.values()), set()

    inc = set()
    exc = set()
    prev = None

    # NOTE: Casting to string and parsing isn't always required, but this
    #   isn't used often enough to require high performance optimization.
    for token in shlex(query):
        if token in MeasurementType.types:
            if prev == "-":
                exc.add(MeasurementType.types[token])
            else:
                inc.add(MeasurementType.types[token])
        elif token not in "+-":
            raise TypeError("Unknown measurement type: {!r}".format(token))
        prev = token
    return inc, exc


def filter_channels(channels: Union[List[Channel], Dict[int, Channel]],
                    measurement_type: Union[MeasurementType, str] = ANY) -> List[Channel]:
    """ Filter a list of `Channel` and/or `SubChannel` instances by their
        measurement type(s).

        :param channels: A list or dictionary of channels/subchannels to filter.
        :param measurement_type: A `MeasurementType`, a measurement type 'key'
            string, or a string of multiple keys generated by adding and/or
            subtracting `MeasurementType` objects. Any 'subtracted' types
            will be excluded.
    """
    # Note: This is separated from `get_channels()` so it can be used
    # elsewhere (i.e., to retrieve channels by type from a `Recorder`)
    if isinstance(channels, dict):
        channels = channels.values()

    if measurement_type == ANY:
        return list(channels)

    # Note: if no `inc`, only `exc` is used
    inc, exc = split_types(measurement_type)

    result = []
    for ch in channels:
        thisType = get_measurement_type(ch)

        if isinstance(thisType, (list, set, tuple)):
            # `ch` is a Channel with SubChannels
            if exc.intersection(thisType):
                # One or more subchannels excluded; exclude channel.
                continue
            elif not inc or inc.intersection(thisType):
                result.append(ch)
                continue

        elif thisType not in exc:
            if not inc or thisType in inc:
                result.append(ch)

    return result



"""
Test configuration import/export.
"""

import os.path
import shutil
import pytest

from endaq.device import getRecorder
import endaq.device.configio

from .fake_recorders import RECORDER_PATHS

# Clear any cached devices, just to be safe
endaq.device.RECORDERS.clear()

# Create parameters, mainly to provide an ID, making the results readable
DEVICES = [pytest.param(getRecorder(path, strict=False), id=os.path.basename(path))
           for path in RECORDER_PATHS]

RECORDER_ROOTS = [os.path.basename(p) for p in RECORDER_PATHS]


@pytest.fixture(scope="session")
def dev_copy_dir(tmpdir_factory):
    """ Fixture to make copies of the fake recorders, for tests that affect
        files on the device (configuration, commands, etc.).
    """
    workdir = tmpdir_factory.mktemp("recorders")
    for src in RECORDER_PATHS:
        dst = workdir / os.path.basename(src)
        shutil.copytree(src, dst)
    return workdir


# ============================================================================
#
# ============================================================================

@pytest.mark.parametrize("path", RECORDER_ROOTS)
def test_basic_export_import(path, dev_copy_dir):
    """ Test basic configuration import/export.
    """
    dev = getRecorder(dev_copy_dir / path, strict=False)
    exportName = dev_copy_dir / f'{dev.partNumber}.xcg'

    originalName = dev.config.name
    # config = dev.config.getConfigValues(defaults=True, none=True)
    endaq.device.configio.exportConfig(dev, exportName)
    assert os.path.exists(exportName)

    foo = endaq.device.configio.deviceFromExport(exportName)
    assert foo.partNumber == dev.partNumber

    dev.config.name = "Changed name"
    assert dev.name != originalName

    endaq.device.configio.importConfig(dev, exportName)
    assert dev.name == originalName

    # TODO: Test more actual config values



"""
Default ConfigUI for LOG-0002-025G-DC data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dVCkAoRYFQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABfpQBZNSZW'
    b'NvcmRpbmcgRGlyZWN0b3J5UAGDFP9/UQSGUkVDT1JEUSSBCFAUk0NvbmZpZ1sweDE2RkY3'
    b'Rl09PTBQFbhEaXJlY3RvcnkgZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5ncy4gOCBjaGFyYW'
    b'N0ZXIgbWF4aW11bUAFQKJQBZVSZWNvcmRpbmcgRmlsZSBQcmVmaXhQAYMV/39RJIEIUBXn'
    b'UHJlZml4IGZvciBzZXJpYWxpemVkIHJlY29yZGluZyBuYW1lcy4gSWYgYmxhbmssIGRldm'
    b'ljZSBzZXJpYWwgbnVtYmVyIHdpbGwgYmUgdXNlZC4gOCBjaGFyYWN0ZXIgbWF4aW11bVAU'
    b'k0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGBAF'
    b'AVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIHRo'
    b'ZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1pZG'
    b'5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZWN0'
    b'b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51bW'
    b'JlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBYpU'
    b'aW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQJlQBYtCdXR0b24gTW9kZVABgxD/f1EBgQ'
    b'BQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3Ig'
    b'YmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBTZW'
    b'NvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BAF0EUUAWOUGx1Zy1JbiBBY3Rpb25QAYMK/39R'
    b'AYEBUBWtU2V0IGFjdGlvbiBmb3IgZGV2aWNlIHdoZW4gcGx1Z2dlZCBpbiB0byBVU0IuQQ'
    b'e1UAWySW1tZWRpYXRlbHkgc3RvcCByZWNvcmRpbmcgYW5kIGFwcGVhciBhcyBVU0IgZHJp'
    b'dmVBB7NQBbBDb21wbGV0ZSByZWNvcmRpbmcgYmVmb3JlIGFwcGVhcmluZyBhcyBVU0IgZH'
    b'JpdmVBB7BQBa1JZ25vcmU6IHN0b3AgcmVjb3JkaW5nIHdoZW4gYnV0dG9uIGlzIHByZXNz'
    b'ZWRBB6VQBaJTdGFydCByZWNvcmRpbmcgd2hlbiBVU0IgY29ubmVjdGVkQCPnUAGDC/9/UB'
    b'XeRW50ZXIgbG9jYWwgdGltZXpvbmUncyBvZmZzZXQgZnJvbSBVVEMgdGltZS4gSW1wYWN0'
    b'cyB0aW1lc3RhbXAgaW5mb3JtYXRpb24gb24gcmVjb3JkaW5nIGZpbGVzLkAvgEAPgEBPgE'
    b'AoSOBQBYhUcmlnZ2Vyc0AHQNpQBYxUcmlnZ2VyIE1vZGVQAYMS/39RAYEAUBVAj0RlbGF5'
    b'IFRoZW4gVHJpZ2dlcjogV2FpdCBzcGVjaWZpZWQgdGltZSwgdGhlbiB3YWl0IGZvciB0cm'
    b'lnZ2VyCkRlbGF5IE9yIFRyaWdnZXI6IEFjcXVpcmUgYWZ0ZXIgd2FpdCBvciBvbiB0cmln'
    b'Z2VyLiBCYXR0ZXJ5IGxpZmUgbWF5IGJlIHJlZHVjZWQuQQeVUAWSRGVsYXkgVGhlbiBUcm'
    b'lnZ2VyQQeTUAWQRGVsYXkgT3IgVHJpZ2dlckAy9FAFjVN0YXJ0IGF0IFRpbWVQAYMP/39Q'
    b'FdtXYWl0IHVudGlsIHNldCB0aW1lIGJlZm9yZSBldmFsdWF0aW5nIG90aGVyIHRyaWdnZX'
    b'IgY29uZGl0aW9ucywgaW5jbHVkaW5nIGEgcmVjb3JkaW5nIGRlbGF5QBGuUAWUUmVjb3Jk'
    b'aW5nIFRpbWUgTGltaXRQAYMN/39RIYT////wUCWHU2Vjb25kc0ARQJBQBZhSZWNvcmRpbm'
    b'cgRGVsYXkvSW50ZXJ2YWxQAYMM/39RAYEAUCWHU2Vjb25kc1EhgwFRgFAV2FRpbWUgYmVm'
    b'b3JlIHN0YXJ0aW5nIGEgcmVjb3JkaW5nLCBhbmQgdGltZSBiZXR3ZWVuIHJlY29yZGluZ3'
    b'MgaWYgJ1JldHJpZ2dlcicgaXMgY2hlY2tlZC5AEa9QBZlSZWNvcmRpbmcgRmlsZSBTaXpl'
    b'IExpbWl0UAGDEf9/USGE////8FAlg0tpQkAQQJdQBYlSZXRyaWdnZXJQAYMO/39QFLBDb2'
    b'5maWdbMHhERkY3Rl09PW51bGwgYW5kIENvbmZpZ1sweDExRkY3Rl09PW51bGxQFc9XaGVu'
    b'IHNldCwgU2xhbSBTdGljayB3aWxsIHJlLWFybSBhZnRlciByZWNvcmRpbmcgdGltZSBvci'
    b'BzaXplIGxpbWl0IGlzIHJlYWNoZWQuQBBAqFAFnldhaXQgZm9yIEFsbCBTZW5zb3IgQ29u'
    b'ZGl0aW9uc1ABgxP/f1AV/ldoZW4gc2V0LCByZWNvcmRpbmcgd2lsbCBvbmx5IHN0YXJ0IH'
    b'doZW4gYWxsIHRyaWdnZXJzIGFyZSBtZXQsIG90aGVyd2lzZSByZWNvcmRpbmcgd2lsbCBz'
    b'dGFydCB3aGVuIGFueSBzZW5zb3IgY29uZGl0aW9uIGlzIG1ldEAYQO5QBZBQcmVzc3VyZS'
    b'BUcmlnZ2VyUAGDBQAkUQCBAECC5VAFlVByZXNzdXJlIFRyaWdnZXIsIExvd1ABgwMAJFEB'
    b'gwFfkFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZS'
    b'BwcmVzc3VyZSB3aW5kb3cuQILmUAWWUHJlc3N1cmUgVHJpZ2dlciwgSGlnaFABgwQAJFEB'
    b'gwGtsFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZS'
    b'BwcmVzc3VyZSB3aW5kb3cuQBhA+1AFk1RlbXBlcmF0dXJlIFRyaWdnZXJQAYMFASRRAIEA'
    b'QGPqUAWYVGVtcGVyYXR1cmUgVHJpZ2dlciwgTG93UAGDAwEkUQKB9lAlgUNREoHYUSKBUF'
    b'AVtlNldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZGUgdGVtcGVyYXR1cmUgd2lu'
    b'ZG93LkBj61AFmVRlbXBlcmF0dXJlIFRyaWdnZXIsIEhpZ2hQAYMEASRRAoEjUCWBQ1ESgd'
    b'hRIoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZSB0ZW1wZXJhdHVy'
    b'ZSB3aW5kb3cuQBhBp1AFmkFjY2VsZXJhdGlvbiBUcmlnZ2VyICgyNWcpUAGDBf8IUBGDB/'
    b'8gUQCBAFAkgFAUlUNvbmZpZ1sweDAxRkYwOF0gPT0gMEBDQKtRE4jAOQAAAAAAAFEjiEA5'
    b'AAAAAAAAUAWZQWNjZWxlcmF0aW9uIFRyaWdnZXIsIExvd1ABgwP/CFECgvmaUBXgSGlnaC'
    b'Bwb3dlciB0cmlnZ2VyIHdpdGggcHJlLXRyaWdnZXIgZGF0YS4gTm90IHJlY29tbWVuZGVk'
    b'IGZvciBsb25nIHdhaXRzIHdpdGhvdXQgZXh0ZXJuYWwgcG93ZXIuUgOIP0kAGQAZABlAQ0'
    b'CsUROIwDkAAAAAAABRI4hAOQAAAAAAAFAFmkFjY2VsZXJhdGlvbiBUcmlnZ2VyLCBIaWdo'
    b'UAGDBP8IUQKCBmZQFeBIaWdoIHBvd2VyIHRyaWdnZXIgd2l0aCBwcmUtdHJpZ2dlciBkYX'
    b'RhLiBOb3QgcmVjb21tZW5kZWQgZm9yIGxvbmcgd2FpdHMgd2l0aG91dCBleHRlcm5hbCBw'
    b'b3dlci5SA4g/SQAZABkAGUAYQZxQBZpBY2NlbGVyYXRpb24gVHJpZ2dlciAoMTZnKVABgw'
    b'f/IFAV1ExvdyBwb3dlciB0cmlnZ2VyIHVzaW5nIGRpZ2l0YWwgYWNjZWxlcm9tZXRlciB3'
    b'aXRoIERDIHJlc3BvbnNlLiBObyBwcmUtdHJpZ2dlciBkYXRhLlEAgQBQEYMF/whQJIBQFJ'
    b'RDb25maWdbMHgxZmYyMF0gPT0gMEBDQLFQBZZBY2NlbGVyYXRpb24gVGhyZXNob2xkUAGD'
    b'BP8gUQGBUFIDiD+wEBAQEBAQUSOIQDAAAAAAAABRE4BQFfJMb3cgcG93ZXIgdHJpZ2dlci'
    b'4gVGhlIG1pbmltdW0gYWNjZWxlcmF0aW9uIChwb3NpdGl2ZSBvciBuZWdhdGl2ZSkgdG8g'
    b'dHJpZ2dlciBhIHJlY29yZGluZy4gTXVzdCBiZSBncmVhdGVyIHRoYW4gMC5AJ8ZQAYMF/y'
    b'BRAYEHQQeRUAWOWCBBeGlzIFRyaWdnZXJBB5FQBY5ZIEF4aXMgVHJpZ2dlckEHkVAFjlog'
    b'QXhpcyBUcmlnZ2VyQA+AQE+AQChGlFAFjE1lYXN1cmVtZW50c0AIQkBQBZcyNWcgQWNjZW'
    b'xlcmF0aW9uIChDaCA4KVAkgEAnQIFRAYEHUAGDAf8IUBWgRW5hYmxlL0Rpc2FibGUgYWNj'
    b'ZWxlcmF0aW9uIGF4ZXNBB5lQBZZFbmFibGUgWCBBeGlzIChDaCA4LjApQQeZUAWWRW5hYm'
    b'xlIFkgQXhpcyAoQ2ggOC4xKUEHmVAFlkVuYWJsZSBaIEF4aXMgKENoIDguMilAEUCaUAWL'
    b'U2FtcGxlIFJhdGVQAYMC/whQFdhTZXQgbWFpbiBhY2NlbGVyYXRpb24gc2FtcGxlIHJhdG'
    b'UgZnJvbSAxMCB0byAyMDAwMCBIei4KSGlnaGVyIHJhdGUgcmVkdWNlcyBiYXR0ZXJ5IGxp'
    b'ZmUuUBSVQ29uZmlnWzB4MDFGRjA4XSA9PSAwUQGCE4hQJYJIelERgQpRIYJOIEARQPxQBa'
    b'NPdmVycmlkZSBBbnRpYWxpYXNpbmcgRmlsdGVyIEN1dG9mZlABgwj/CFEBggPoUCWCSHpR'
    b'IYITiFAVQKVEZWZhdWx0LCB3aGVuIG5vdCBjaGVja2VkLCBvZiBvbmUtZmlmdGggc2FtcG'
    b'xlIGZyZXF1ZW5jeS4gQ3V0LW9mZiBmcmVxdWVuY3kgaGFzIDI5JSBhdHRlbnVhdGlvbi4K'
    b'NXRoIG9yZGVyIEJ1dHRlcndvcnRoLiBTb21lIGF0dGVudWF0aW9uIGF0IDYwJSBvZiBjdX'
    b'Qtb2ZmIGZyZXF1ZW5jeS5QFJVDb25maWdbMHgwMUZGMDhdID09IDBACEC6UAWbMTZnIERD'
    b'IEFjY2VsZXJhdGlvbiAoQ2ggMzIpQCefUAGDAf8gUQGBAUEHklAFj0VuYWJsZSBBbGwgQX'
    b'hlc0AX91AFi1NhbXBsZSBSYXRlUAGDAv8gUQGCAZBQJYJIelAUk25vdCBDb25maWdbMHgx'
    b'ZmYyMF1BB4VRAYIMgEEHhVEBggZAQQeFUQGCAyBBB4VRAYIBkEEHhFEBgchBB4RRAYFkQQ'
    b'eEUQGBMkEHhFEBgRlBB4RRAYEMQAhBMFAFqENvbnRyb2wgUGFkIFRlbXBlcmF0dXJlL1By'
    b'ZXNzdXJlIChDaCA1OSlAJ0CbUQGBA1ABgwH/O1AVpUVuYWJsZS9EaXNhYmxlIHNlbnNvcn'
    b'Mgb24gY29udHJvbCBwYWRBB5xQBZlFbmFibGUgUHJlc3N1cmUgKENoIDU5LjApQQefUAWc'
    b'RW5hYmxlIFRlbXBlcmF0dXJlIChDaCA1OS4xKUEHpVAFokVuYWJsZSBSZWxhdGl2ZSBIdW'
    b'1pZGl0eSAoQ2ggNTkuMilAEeNQBYtTYW1wbGUgUmF0ZVABgwL/O1AUlENvbmZpZ1sweDFm'
    b'ZjNCXSA9PSAwUBWkU2V0IHNhbXBsZSBmcmVxdWVuY3kgZnJvbSAxIHRvIDEwIEh6UQGBCl'
    b'Algkh6URGBAVEhgQpACEHJUAWZSW5lcnRpYWwgTWVhc3VyZW1lbnQgVW5pdEAHQJJQBZBB'
    b'Y3F1aXNpdGlvbiBNb2RlUAGDAf8rUQGBEEEHilAFg09mZlEBgQBBB6hQBaFBYnNvbHV0ZS'
    b'BPcmllbnRhdGlvbiAoUXVhdGVybmlvbilRAYEIQQeoUAWhUmVsYXRpdmUgT3JpZW50YXRp'
    b'b24gKFF1YXRlcm5pb24pUQGBEEEHj1AFiFJvdGF0aW9uUQGBAkABQIZQBZVPcmllbnRhdG'
    b'lvbiBEYXRhIFJhdGVQAYMCCCtRAYFkUCWCSHpREYEBUSGByFAVvVNhbXBsZSByYXRlIGZv'
    b'ciBkaXJlY3Rpb25hbCBkYXRhIGZyb20gMSB0byAyMDAgSHouIENoYW5uZWwgNjVQFJRDb2'
    b'5maWdbMHgwMUZGMkJdIDwgOEABQIlQBZJSb3RhdGlvbiBEYXRhIFJhdGVQAYMCAitRAYFk'
    b'UCWCSHpREYEBUSGByFAVvlNldCBzYW1wbGUgcmF0ZSBmb3Igcm90YXRpb24gZGF0YSBmcm'
    b'9tIDEgdG8gMjAwIEh6LiBDaGFubmVsIDQ3UBSZQ29uZmlnWzB4MDFGRjJCXSAmIDIgPT0g'
    b'MEAI+VAFoENvbnRyb2wgUGFkIExpZ2h0IFNlbnNvciAoQ2ggNzYpQCfTUQGBAVABgwH/TF'
    b'AVo0VuYWJsZS9EaXNhYmxlIGxpZ2h0IHNlbnNvciBhdCA0IEh6QQegUAWdRW5hYmxlIExp'
    b'Z2h0IFNlbnNvciAoQ2ggNzYuMClAD4BAT4BKCIBKKIBKSIA='
)



"""
Default ConfigUI for Sx-D8 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'GkXfo6lChoEBQveBAULygQRC84EIQoKObWlkZS5zcy5jb25maWdCh4ECQoWBAnd3EAAYIk'
    b'AoIAa3UAWHR2VuZXJhbEAFQHNQBYtEZXZpY2UgTmFtZVABgwj/f1EkgUBQFdhBIGN1c3Rv'
    b'bSBuYW1lIGZvciB0aGUgcmVjb3JkZXIuIE5vdCB0aGUgc2FtZSBhcyB0aGUgdm9sdW1lIG'
    b'xhYmVsLiA2NCBjaGFyYWN0ZXJzIG1heGltdW0uQAVAc1AFjERldmljZSBOb3Rlc1ABgwn/'
    b'f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24sIH'
    b'VzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABUCZUAWVQ3Vz'
    b'dG9tIFJlY29yZGluZyBUYWdzUAGDF/9/UQSAUSSBQFAV8UtleXdvcmRzIGZvciBvcmdhbm'
    b'l6aW5nIGFuZCBpZGVudGlmeWluZyByZWNvcmRpbmdzLCBzZXBhcmF0ZWQgYnkgY29tbWEg'
    b'b3Igc3BhY2UuIExpbWl0IGlzIDIwIHRhZ3MsIDY0IGNoYXJhY3RlcnMuQAdBPFAFj0ZpbG'
    b'UgTmFtZSBTdHlsZVABgxb/f1EBgQFQFUD8VGltZS1CYXNlZCBmaWxlIG5hbWVzIHVzZSBh'
    b'IGRpcmVjdG9yeSBiYXNlZCBvbiB0aGUgZGF0ZSwgYW5kIGZpbGUgbmFtZSBiYXNlZCBvbi'
    b'B0aGUgc2Vjb25kcyBzaW5jZSBtaWRuaWdodC4KU2VyaWFsaXplZCBmaWxlIG5hbWVzIHVz'
    b'ZSB0aGUgc3BlY2lmaWVkIGRpcmVjdG9yeSBhbmQgZmlsZSBuYW1lIHByZWZpeCwgZm9sbG'
    b'93ZWQgYnkgYSBudW1iZXIuIApOdW1iZXIgaXMgcmVzZXQgd2hlbiBkaXJlY3Rvcnkgb3Ig'
    b'cHJlZml4IGNoYW5nZXMuQQeNUAWKVGltZSBCYXNlZEEHjVAFilNlcmlhbGl6ZWRABUB7UA'
    b'WTUmVjb3JkaW5nIERpcmVjdG9yeVABgxT/f1EEhlJFQ09SRFEkgRBQFJNDb25maWdbMHgx'
    b'NkZGN0ZdPT0wUBW5RGlyZWN0b3J5IGZvciBzZXJpYWxpemVkIHJlY29yZGluZ3MuIDE2IG'
    b'NoYXJhY3RlciBtYXhpbXVtQAVAo1AFlVJlY29yZGluZyBGaWxlIFByZWZpeFABgxX/f1Ek'
    b'gRBQFehQcmVmaXggZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5nIG5hbWVzLiBJZiBibGFuay'
    b'wgZGV2aWNlIHNlcmlhbCBudW1iZXIgd2lsbCBiZSB1c2VkLiAxNiBjaGFyYWN0ZXIgbWF4'
    b'aW11bVAUk0NvbmZpZ1sweDE2RkY3Rl09PTBAF0EUUAWOUGx1Zy1JbiBBY3Rpb25QAYMK/3'
    b'9RAYEBUBWtU2V0IGFjdGlvbiBmb3IgZGV2aWNlIHdoZW4gcGx1Z2dlZCBpbiB0byBVU0Iu'
    b'QQe1UAWySW1tZWRpYXRlbHkgc3RvcCByZWNvcmRpbmcgYW5kIGFwcGVhciBhcyBVU0IgZH'
    b'JpdmVBB7NQBbBDb21wbGV0ZSByZWNvcmRpbmcgYmVmb3JlIGFwcGVhcmluZyBhcyBVU0Ig'
    b'ZHJpdmVBB7BQBa1JZ25vcmU6IHN0b3AgcmVjb3JkaW5nIHdoZW4gYnV0dG9uIGlzIHByZX'
    b'NzZWRBB6VQBaJTdGFydCByZWNvcmRpbmcgd2hlbiBVU0IgY29ubmVjdGVkQAdBLFAFi0J1'
    b'dHRvbiBNb2RlUAGDEP9/UQGBAFAVQMFPbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZS'
    b'AzLXNlY29uZCBwcmVzcyBmb3IgYmF0dGVyeSBjaGVjay4KTk9URTogSWYgJ1N0YXJ0LU9u'
    b'bHknIGlzIHNlbGVjdGVkIGFuZCBubyByZWNvcmRpbmcgdGltZSBsaW1pdCBvciBtYXggZm'
    b'lsZSBzaXplIGlzIHNldCwgcmVjb3JkaW5ncyBjYW4gYmUgc3RvcHBlZCB3aXRoIFVTQiBw'
    b'bHVnLWluQQeXUAWUSW1tZWRpYXRlIFN0YXJ0L1N0b3BBB6JQBZ9PbmUgU2Vjb25kIFByZX'
    b'NzIGZvciBTdGFydC9TdG9wQQeNUAWKU3RhcnQtT25seUAjQGdQAYML/39QFd5FbnRlciBs'
    b'b2NhbCB0aW1lem9uZSdzIG9mZnNldCBmcm9tIFVUQyB0aW1lLiBJbXBhY3RzIHRpbWVzdG'
    b'FtcCBpbmZvcm1hdGlvbiBvbiByZWNvcmRpbmcgZmlsZXMuQC+AQA+AQE+AQCggCoNQBYhU'
    b'cmlnZ2Vyc0AHQNpQBYxUcmlnZ2VyIE1vZGVQAYMS/39RAYEAUBVAj0RlbGF5IFRoZW4gVH'
    b'JpZ2dlcjogV2FpdCBzcGVjaWZpZWQgdGltZSwgdGhlbiB3YWl0IGZvciB0cmlnZ2VyCkRl'
    b'bGF5IE9yIFRyaWdnZXI6IEFjcXVpcmUgYWZ0ZXIgd2FpdCBvciBvbiB0cmlnZ2VyLiBCYX'
    b'R0ZXJ5IGxpZmUgbWF5IGJlIHJlZHVjZWQuQQeVUAWSRGVsYXkgVGhlbiBUcmlnZ2VyQQeT'
    b'UAWQRGVsYXkgT3IgVHJpZ2dlckAyQHRQBY1TdGFydCBhdCBUaW1lUAGDD/9/UBXbV2FpdC'
    b'B1bnRpbCBzZXQgdGltZSBiZWZvcmUgZXZhbHVhdGluZyBvdGhlciB0cmlnZ2VyIGNvbmRp'
    b'dGlvbnMsIGluY2x1ZGluZyBhIHJlY29yZGluZyBkZWxheUARQJZQBZxQcmUtUmVjb3JkaW'
    b'5nIERlbGF5L0ludGVydmFsUAGDDP9/UQGBAFAlh1NlY29uZHNRIYMBUYBQFdpEZWxheSBi'
    b'ZWZvcmUgc3RhcnRpbmcgYSByZWNvcmRpbmcsIGluY2x1ZGluZyBiZXR3ZWVuIHJlY29yZG'
    b'luZ3MgaWYgJ1JldHJpZ2dlcicgaXMgY2hlY2tlZC5AEUAuUAWUUmVjb3JkaW5nIFRpbWUg'
    b'TGltaXRQAYMN/39RIYT////wUCWHU2Vjb25kc0ARQC9QBZlSZWNvcmRpbmcgRmlsZSBTaX'
    b'plIExpbWl0UAGDEf9/USGE////8FAlg0tpQkAQQJJQBYlSZXRyaWdnZXJQAYMO/39QFLBD'
    b'b25maWdbMHhERkY3Rl09PW51bGwgYW5kIENvbmZpZ1sweDExRkY3Rl09PW51bGxQFcpXaG'
    b'VuIHNldCwgZW5EQVEgd2lsbCByZS1hcm0gYWZ0ZXIgcmVjb3JkaW5nIHRpbWUgb3Igc2l6'
    b'ZSBsaW1pdCBpcyByZWFjaGVkLkAQQKhQBZ5XYWl0IGZvciBBbGwgU2Vuc29yIENvbmRpdG'
    b'lvbnNQAYMT/39QFf5XaGVuIHNldCwgcmVjb3JkaW5nIHdpbGwgb25seSBzdGFydCB3aGVu'
    b'IGFsbCB0cmlnZ2VycyBhcmUgbWV0LCBvdGhlcndpc2UgcmVjb3JkaW5nIHdpbGwgc3Rhcn'
    b'Qgd2hlbiBhbnkgc2Vuc29yIGNvbmRpdGlvbiBpcyBtZXRAGEELUAWZQ3RybCBQYWQgUHJl'
    b'c3N1cmUgVHJpZ2dlclABgwUAJ1EAgQBAgkBuUAWeQ3RybCBQYWQgUHJlc3N1cmUgVHJpZ2'
    b'dlciwgTG93UAGDAwAnUQGDAV+QUCWCUGFRIYMB1MBQFbNTZXQgdG8gc3RhcnQgc2FtcGxp'
    b'bmcgd2hlbiBvdXRzaWRlIHByZXNzdXJlIHdpbmRvdy5AgkBvUAWfQ3RybCBQYWQgUHJlc3'
    b'N1cmUgVHJpZ2dlciwgSGlnaFABgwQAJ1EBgwGtsFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0'
    b'YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZSBwcmVzc3VyZSB3aW5kb3cuQBhBdlAFmUludG'
    b'VybmFsIFByZXNzdXJlIFRyaWdnZXJQAYMFACRRAIEAUBToKENvbmZpZ1sweDFmZjE0XSBh'
    b'bmQgKENvbmZpZ1sweDFmZjE0XSAgJiAxID09IDApKSBvciAoQ29uZmlnWzB4MDFGRjI0XS'
    b'BhbmQgKENvbmZpZ1sweDAxRkYyNF0gICYgMSA9PSAwKSlAgkBuUAWeSW50ZXJuYWwgUHJl'
    b'c3N1cmUgVHJpZ2dlciwgTG93UAGDAwAkUQGDAV+QUCWCUGFRIYMB1MBQFbNTZXQgdG8gc3'
    b'RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaWRlIHByZXNzdXJlIHdpbmRvdy5AgkBvUAWfSW50'
    b'ZXJuYWwgUHJlc3N1cmUgVHJpZ2dlciwgSGlnaFABgwQAJFEBgwGtsFAlglBhUSGDAdTAUB'
    b'WzU2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZSBwcmVzc3VyZSB3aW5kb3cu'
    b'QBhBGFAFnEN0cmwgUGFkIFRlbXBlcmF0dXJlIFRyaWdnZXJQAYMFASdRAIEAQGNAc1AFoU'
    b'N0cmwgUGFkIFRlbXBlcmF0dXJlIFRyaWdnZXIsIExvd1ABgwMBJ1ECgfZQJYFDURKB2FEi'
    b'gVBQFbZTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaWRlIHRlbXBlcmF0dXJlIH'
    b'dpbmRvdy5AY0B0UAWiQ3RybCBQYWQgVGVtcGVyYXR1cmUgVHJpZ2dlciwgSGlnaFABgwQB'
    b'J1ECgSNQJYFDURKB2FEigVBQFbZTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaW'
    b'RlIHRlbXBlcmF0dXJlIHdpbmRvdy5AGEGDUAWcSW50ZXJuYWwgVGVtcGVyYXR1cmUgVHJp'
    b'Z2dlclABgwUBJFEAgQBQFOgoQ29uZmlnWzB4MWZmMTRdIGFuZCAoQ29uZmlnWzB4MWZmMT'
    b'RdICAmIDIgPT0gMCkpIG9yIChDb25maWdbMHgwMUZGMjRdIGFuZCAoQ29uZmlnWzB4MDFG'
    b'RjI0XSAgJiAyID09IDApKUBjQHNQBaFJbnRlcm5hbCBUZW1wZXJhdHVyZSBUcmlnZ2VyLC'
    b'BMb3dQAYMDASRRAoH2UCWBQ1ESgdhRIoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdo'
    b'ZW4gb3V0c2lkZSB0ZW1wZXJhdHVyZSB3aW5kb3cuQGNAdFAFokludGVybmFsIFRlbXBlcm'
    b'F0dXJlIFRyaWdnZXIsIEhpZ2hQAYMEASRRAoEjUCWBQ1ESgdhRIoFQUBW2U2V0IHRvIHN0'
    b'YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZSB0ZW1wZXJhdHVyZSB3aW5kb3cuQBhBq1AFmU'
    b'FjY2VsZXJhdGlvbiBUcmlnZ2VyICg4ZylQAYMH/1BQFdRMb3cgcG93ZXIgdHJpZ2dlciB1'
    b'c2luZyBkaWdpdGFsIGFjY2VsZXJvbWV0ZXIgd2l0aCBEQyByZXNwb25zZS4gTm8gcHJlLX'
    b'RyaWdnZXIgZGF0YS5RAIEAUBGDBf8IUBGDB/8gUCSAUBSUQ29uZmlnWzB4MWZmNTBdID09'
    b'IDBAQ0C6UAWWQWNjZWxlcmF0aW9uIFRocmVzaG9sZFABgwT/UFEBgiAAUgOIPyAAEAAQAB'
    b'BRI4hAIAAAAAAAAFETiAAAAAAAAAAAUBXyTG93IHBvd2VyIHRyaWdnZXIuIFRoZSBtaW5p'
    b'bXVtIGFjY2VsZXJhdGlvbiAocG9zaXRpdmUgb3IgbmVnYXRpdmUpIHRvIHRyaWdnZXIgYS'
    b'ByZWNvcmRpbmcuIE11c3QgYmUgZ3JlYXRlciB0aGFuIDAuQCdARlABgwX/UFEBgQdBB5FQ'
    b'BY5YIEF4aXMgVHJpZ2dlckEHkVAFjlkgQXhpcyBUcmlnZ2VyQQeRUAWOWiBBeGlzIFRyaW'
    b'dnZXJAD4BAT4BAKCAG0FAFjE1lYXN1cmVtZW50c0AIQRxQBZo4ZyBEQyBBY2NlbGVyYXRp'
    b'b24gKENoIDgwKUAnQB9QAYMB/1BRAYEBQQeSUAWPRW5hYmxlIEFsbCBBeGVzQBdA2FAFi1'
    b'NhbXBsZSBSYXRlUAGDAv9QUQGCAfRQJYJIelAV3kxvdyBwYXNzIGZpbHRlciBpcyBzZXQg'
    b'dG8gMjUlIG9mIGRhdGEgcmF0ZSwgc28gYSBkYXRhIHJhdGUgb2YgMjAwMEh6IGhhcyBhIE'
    b'xQIGZpbHRlciBhdCA1MDAgSHpQFJNub3QgQ29uZmlnWzB4MWZmNTBdQQeFUQGCD6BBB4VR'
    b'AYIH0EEHhVEBggPoQQeFUQGCAfRBB4RRAYH6QQeEUQGBfUEHhFEBgT9BB4RRAYEgQQeEUQ'
    b'GBEEAIQdVQBaFCSEkxNjBCIEluZXJ0aWFsIE1lYXN1cmVtZW50IFVuaXRAB0CWUAWQQWNx'
    b'dWlzaXRpb24gTW9kZVABgwH/K1AlgSBRAYEQQQeKUAWDT2ZmUQGBAEEHqFAFoUFic29sdX'
    b'RlIE9yaWVudGF0aW9uIChRdWF0ZXJuaW9uKVEBgQhBB6hQBaFSZWxhdGl2ZSBPcmllbnRh'
    b'dGlvbiAoUXVhdGVybmlvbilRAYEQQQePUAWIUm90YXRpb25RAYECQAFAhlAFlU9yaWVudG'
    b'F0aW9uIERhdGEgUmF0ZVABgwIIK1EBgWRQJYJIelERgQFRIYHIUBW9U2FtcGxlIHJhdGUg'
    b'Zm9yIGRpcmVjdGlvbmFsIGRhdGEgZnJvbSAxIHRvIDIwMCBIei4gQ2hhbm5lbCA2NVAUlE'
    b'NvbmZpZ1sweDAxRkYyQl0gPCA4QAFAiVAFklJvdGF0aW9uIERhdGEgUmF0ZVABgwICK1EB'
    b'gWRQJYJIelERgQFRIYHIUBW+U2V0IHNhbXBsZSByYXRlIGZvciByb3RhdGlvbiBkYXRhIG'
    b'Zyb20gMSB0byAyMDAgSHouIENoYW5uZWwgNDdQFJlDb25maWdbMHgwMUZGMkJdICYgMiA9'
    b'PSAwQAhBSFAFsUNvbnRyb2wgUGFkIFRlbXBlcmF0dXJlL1ByZXNzdXJlL0h1bWlkaXR5IC'
    b'hDaCA1OSlAJ0CgUQGBA1ABgwH/O1AVqkVuYWJsZS9EaXNhYmxlIENvbnRyb2wgUGFkIFNl'
    b'bnNvciBDaGFubmVsc0EHnFAFmUVuYWJsZSBQcmVzc3VyZSAoQ2ggNTkuMClBB59QBZxFbm'
    b'FibGUgVGVtcGVyYXR1cmUgKENoIDU5LjEpQQelUAWiRW5hYmxlIFJlbGF0aXZlIEh1bWlk'
    b'aXR5IChDaCA1OS4yKUARQGxQBZRDdHJsIFBhZCBTYW1wbGUgUmF0ZVABgwL/O1AUlENvbm'
    b'ZpZ1sweDFmZjNCXSA9PSAwUBWkU2V0IHNhbXBsZSBmcmVxdWVuY3kgZnJvbSAxIHRvIDEw'
    b'IEh6UQGBClAlgkh6URGBAVEhgQpACEFCUAWuSW50ZXJuYWwgVGVtcGVyYXR1cmUvUHJlc3'
    b'N1cmUvSHVtaWRpdHkgKENoIDIwKUAnQJ1RAYEDUAGDAf8UUBWnRW5hYmxlL0Rpc2FibGUg'
    b'SW50ZXJuYWwgU2Vuc29yIENoYW5uZWxzQQecUAWZRW5hYmxlIFByZXNzdXJlIChDaCAyMC'
    b'4wKUEHn1AFnEVuYWJsZSBUZW1wZXJhdHVyZSAoQ2ggMjAuMSlBB6VQBaJFbmFibGUgUmVs'
    b'YXRpdmUgSHVtaWRpdHkgKENoIDIwLjIpQBFAbFAFlEludGVybmFsIFNhbXBsZSBSYXRlUA'
    b'GDAv8UUBSUQ29uZmlnWzB4MWZmMTRdID09IDBQFaRTZXQgc2FtcGxlIGZyZXF1ZW5jeSBm'
    b'cm9tIDEgdG8gMTAgSHpRAYEKUCWCSHpREYEBUSGBCkAIQHZQBaBDb250cm9sIFBhZCBMaW'
    b'dodCBTZW5zb3IgKENoIDc2KUAnQE9QAYMB/0xQFaNFbmFibGUvRGlzYWJsZSBsaWdodCBz'
    b'ZW5zb3IgYXQgNCBIekEHoFAFnUVuYWJsZSBMaWdodCBTZW5zb3IgKENoIDc2LjApQAhAsl'
    b'AFmEJNSTI3MCBHeXJvc2NvcGUgKENoIDg0KUAnQB9QAYMB/1RRAYEAQQeSUAWPRW5hYmxl'
    b'IEFsbCBBeGVzQBdAcFAFi1NhbXBsZSBSYXRlUAGDAv9UUQGCAZBQJYJIelAUk25vdCBDb2'
    b'5maWdbMHgxZmY1NF1BB4VRAYIMgEEHhVEBggZAQQeFUQGCAyBBB4VRAYIBkEEHhFEBgchB'
    b'B4RRAYFkQQeEUQGBMkEHhFEBgRlAD4BAT4BKCIBKKIBKSIA='
)



"""
Conversion between legacy (pre EFM32 GG0 FWRev 15) and contemporary config
file formats. These get used automatically for older SlamStick devices.
"""

__author__ = "dstokes"
__copyright__ = "Copyright 2023 Mide Technology Corporation"

from typing import Any, AnyStr, Dict, TYPE_CHECKING, Union

from ebmlite import loadSchema, Document, MasterElement

import logging
logger = logging.getLogger('endaq.device.legacy')

from .measurement import ACCELERATION

if TYPE_CHECKING:
    from .base import Recorder
    from .config import ConfigInterface

# ==============================================================================
#
# ==============================================================================


def _copyItems(oldD: AnyStr, newD: AnyStr, *keyPairs):
    """ Utility function to copy existing, non-`None` items from one dictionary
        to another, using different keys.
    """
    for oldK, newK in keyPairs:
        val = oldD.get(oldK)
        if val is not None and val != "":
            newD[newK] = val


def convertConfigItems(config: dict,
                       device: "Recorder") -> Dict[int, Any]:
    """ Convert legacy-style ``RecorderConfiguration`` items to the
        new ``RecorderConfigurationList`` style.

        :param config: A dictionary, dumped from an EBML
            ``RecorderConfiguration`` element.
        :param device: The device containing the data.
        :return: A dictionary of ConfigIDs and values (the 'new' format).
    """
    if not config:
        return {}

    newData = {}
    channels = device.getChannels()

    if 'RecorderConfiguration' in config:
        config = config['RecorderConfiguration']

    # Combine 'root' dictionaries for easy access
    basicConfig = config.get('SSXBasicRecorderConfiguration', {})
    userConfig = config.get('RecorderUserData', {})
    triggerConfig = config.get('SSXTriggerConfiguration', {})
    channelConfig = config.get('SSXChannelConfiguration', [])

    # Basic stuff. Items only added if they exist in the old config data.
    _copyItems(basicConfig, newData,
               ('SampleFreq', 0x02ff08),
               ('AAFilterCornerFreq', 0x08ff08),
               ('PlugPolicy', 0x0aff7f),
               ('UTCOffset', 0x0bff7f))

    _copyItems(userConfig, newData,
               ('RecorderName', 0x08ff7f),
               ('RecorderDesc', 0x09ff7f))

    _copyItems(triggerConfig, newData,
               ('WakeTimeUTC', 0x0fff7f),
               ('PreRecordDelay', 0x0cff7f),
               ('RecordingTime', 0x0dff7f),
               ('AutoRearm', 0x0eff7f))

    # Channel configuration.
    for ch in channelConfig:
        chId = ch.get('ConfigChannel')
        enables = ch.get('SubChannelEnableMap', 0xFF)
        sampFreq = ch.get('ChannelSampleFreq')

        if chId is None:
            continue
        if enables is not None:
            newData[0x01FF00 | (chId & 0xFF)] = enables
        if sampFreq is not None:
            newData[0x02FF00 | (chId & 0xFF)] = sampFreq

    # Trigger configuration.
    dcAccelMap = 0  # For building DC accelerometer's 'participation map'.

    for trigger in triggerConfig.get('Trigger', []):
        chId = trigger.get('TriggerChannel')
        subchId = trigger.get('TriggerSubChannel', 0xff) & 0xff
        trigLo = trigger.get('TriggerWindowLo')
        trigHi = trigger.get('TriggerWindowHi')

        if chId is None:
            continue

        if chId not in channels:
            # Convert very early firmware channel IDs
            if chId == 0 and 8 in channels:
                chId = 8
            elif chId == 1 and 36 in channels:
                chId = 36
            else:
                logger.warning("Ignoring bad trigger channel ID: %r" % chId)
                continue

        combinedId = (subchId << 8) | (chId & 0xFF)

        if chId == 32:
            # Special case: DC accelerometer (new data uses 'participation map'
            # instead of individual subchannel triggers).
            dcAccelMap |= (1 << subchId)
            combinedId = combinedId | 0x00FF00
        else:
            newData[0x050000 | combinedId] = 1

        if trigLo is not None:
            newData[0x030000 | combinedId] = trigLo
        if trigHi is not None:
            newData[0x040000 | combinedId] = trigHi

    if dcAccelMap > 0:
        newData[0x05FF20] = dcAccelMap

    accels = device.getChannels(ACCELERATION)
    analogChannel = accels.get(8, accels.get(0, None))

    if analogChannel:
        enableId = 0x01ff00 | (analogChannel.id & 0xFF)
        if enableId not in newData:
            newData[enableId] = 0
            for _ch in analogChannel.subchannels:
                newData[enableId] = (newData[enableId] << 1) | 1

    return newData


def convertConfigFile(original: Union[Document, MasterElement],
                      interface: "ConfigInterface") -> MasterElement:
    """ Convert legacy-style ``RecorderConfiguration`` data to the
        new ``RecorderConfigurationList`` style.

        :param original: The original, legacy format EBML config document.
        :param interface: The `ConfigInterface` requesting the conversion.
        :return: An EBML config document in the new format.
    """
    vals = convertConfigItems(original.dump(), interface.device)

    data = []
    items = {'RecorderConfigurationItem': data}

    for k, v in vals.items():
        if k in interface._items:
            data.append({'ConfigID': k,
                         interface._items[k].vtype: v})

    schema = loadSchema('mide_ide.xml')
    ebml = schema.encodes({'RecorderConfigurationList': items},
                          headers=False)
    return schema.loads(ebml)


def generateLegacyConfig(configData: dict, device: "Recorder") -> dict:
    """ Build an EBML-encodable set of nested dictionaries containing the
        dialog's configuration values in the legacy format. Note: the
        `configData` should not contain any `None` values; these will be
        ignored.

        :param configData: A dictionary of configuration data, in the new
            style (flat, keyed by ConfigID).
        :param device: The device to which to write.
    """
    # Copy the data, just in case.
    configData = configData.copy()

    # Individual dictionaries/lists for each section of the old config
    userConfig = {}
    basicConfig = {}
    triggerConfig = {}
    channelConfig = []

    # Basic stuff. Items only added if they exist in the new config data.
    _copyItems(configData, userConfig,
               (0x08ff7f, 'RecorderName'),
               (0x09ff7f, 'RecorderDesc'))

    _copyItems(configData, basicConfig,
               (0x02ff08, 'SampleFreq'),
               (0x08ff08, 'AAFilterCornerFreq'),
               (0x0aff7f, 'PlugPolicy'),
               (0x0bff7f, 'UTCOffset'))

    _copyItems(configData, triggerConfig,
               (0x0fff7f, 'WakeTimeUTC'),
               (0x0cff7f, 'PreRecordDelay'),
               (0x0dff7f, 'RecordingTime'),
               (0x0eff7f, 'AutoRearm'))

    # Trigger configuration: separate master elements for each subchannel.
    triggers = []

    # Get all trigger enables/subchannel participation maps
    for t in [k for k in configData if (k & 0xFF0000 == 0x050000)]:
        combinedId = t & 0x00FFFF
        trigLo = configData.get(0x030000 | combinedId)
        trigHi = configData.get(0x040000 | combinedId)

        trig = {'TriggerChannel': combinedId & 0xFF}

        # Special case: DC accelerometer, which uses a 'participation' bitmap
        # instead of having explicit ConfigID items for each subchannel. In
        # legacy data, each subchannel has its own trigger element, with the
        # same threshold value in each (can't change per subchannel).
        if t == 0x05FF20:
            v = configData[0x05FF20]
            trigHi = configData.get(0x04ff20)
            if trigHi is not None:
                trig['TriggerWindowHi'] = trigHi
            for i in range(3):
                if (v >> i) & 1:
                    d = trig.copy()
                    d['TriggerSubChannel'] = i
                    triggers.append(d)
        else:
            trig['TriggerSubChannel'] = ((combinedId & 0x00FF00) >> 8)
            triggers.append(trig)

        if trigLo is not None:
            trig['TriggerWindowLo'] = trigLo
        if trigHi is not None:
            trig['TriggerWindowHi'] = trigHi

    if triggers:
        triggerConfig['Trigger'] = triggers

    # Channel configuration: per-axis enables, sample rate for some.
    for c in device.getChannels():
        combinedId = 0xFF00 | (c & 0xFF)
        d = {}
        _copyItems(configData, d,
                   (0x020000 | combinedId, "ChannelSampleFreq"),
                   (0x010000 | combinedId, "SubChannelEnableMap"))

        # Only save if something's been set.
        if d:
            d['ConfigChannel'] = c
            channelConfig.append(d)

    # Build the complete old-style configuration dictionary. Only add stuff
    # with content.
    legacyConfigData = {}

    if basicConfig:
        legacyConfigData['SSXBasicRecorderConfiguration'] = basicConfig
    if userConfig:
        legacyConfigData['RecorderUserData'] = userConfig
    if triggerConfig:
        legacyConfigData['SSXTriggerConfiguration'] = triggerConfig
    if channelConfig:
        legacyConfigData['SSXChannelConfiguration'] = channelConfig

    return {'RecorderConfiguration': legacyConfigData}



"""
Default ConfigUI for W8-E100D40 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dZHEAoSDxQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABUCZUAWVQ3'
    b'VzdG9tIFJlY29yZGluZyBUYWdzUAGDF/9/UQSAUSSBQFAV8UtleXdvcmRzIGZvciBvcmdh'
    b'bml6aW5nIGFuZCBpZGVudGlmeWluZyByZWNvcmRpbmdzLCBzZXBhcmF0ZWQgYnkgY29tbW'
    b'Egb3Igc3BhY2UuIExpbWl0IGlzIDIwIHRhZ3MsIDY0IGNoYXJhY3RlcnMuQAX7UAWTUmVj'
    b'b3JkaW5nIERpcmVjdG9yeVABgxT/f1EEhlJFQ09SRFEkgRBQFJNDb25maWdbMHgxNkZGN0'
    b'ZdPT0wUBW5RGlyZWN0b3J5IGZvciBzZXJpYWxpemVkIHJlY29yZGluZ3MuIDE2IGNoYXJh'
    b'Y3RlciBtYXhpbXVtQAVAo1AFlVJlY29yZGluZyBGaWxlIFByZWZpeFABgxX/f1EkgRBQFe'
    b'hQcmVmaXggZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5nIG5hbWVzLiBJZiBibGFuaywgZGV2'
    b'aWNlIHNlcmlhbCBudW1iZXIgd2lsbCBiZSB1c2VkLiAxNiBjaGFyYWN0ZXIgbWF4aW11bV'
    b'AUk0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGB'
    b'AVAVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIH'
    b'RoZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1p'
    b'ZG5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZW'
    b'N0b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51'
    b'bWJlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBY'
    b'pUaW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQKlQBYtCdXR0b24gTW9kZVABgxD/f1EB'
    b'gQBQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3'
    b'IgYmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBT'
    b'ZWNvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BBB41QBYpTdGFydC1Pbmx5QBdBFFAFjlBsdW'
    b'ctSW4gQWN0aW9uUAGDCv9/UQGBAVAVrVNldCBhY3Rpb24gZm9yIGRldmljZSB3aGVuIHBs'
    b'dWdnZWQgaW4gdG8gVVNCLkEHtVAFskltbWVkaWF0ZWx5IHN0b3AgcmVjb3JkaW5nIGFuZC'
    b'BhcHBlYXIgYXMgVVNCIGRyaXZlQQezUAWwQ29tcGxldGUgcmVjb3JkaW5nIGJlZm9yZSBh'
    b'cHBlYXJpbmcgYXMgVVNCIGRyaXZlQQewUAWtSWdub3JlOiBzdG9wIHJlY29yZGluZyB3aG'
    b'VuIGJ1dHRvbiBpcyBwcmVzc2VkQQelUAWiU3RhcnQgcmVjb3JkaW5nIHdoZW4gVVNCIGNv'
    b'bm5lY3RlZEAj51ABgwv/f1AV3kVudGVyIGxvY2FsIHRpbWV6b25lJ3Mgb2Zmc2V0IGZyb2'
    b'0gVVRDIHRpbWUuIEltcGFjdHMgdGltZXN0YW1wIGluZm9ybWF0aW9uIG9uIHJlY29yZGlu'
    b'ZyBmaWxlcy5AL4BAGEIIUAWMRW5hYmxlIFdpLUZpUAGDGP9/UQGBAUAn7FABgxn/f1EBgQ'
    b'FBB6lQBaZVcGxvYWQgcmVjb3JkaW5nIHRvIGNsb3VkIGFmdGVyIGZpbmlzaFAVs1NlbGVj'
    b'dCB0byB1cGxvYWQgZWFjaCBmaWxlIHdoZW4gcmVjb3JkaW5nIGNvbXBsZXRlc0AnQLZQAY'
    b'Ma/39RAYEAQQejUAWgRGVsZXRlIGZyb20gc3RvcmFnZSBhZnRlciB1cGxvYWRQFUCCUmVj'
    b'b3JkaW5nIGZpbGUgd2lsbCBiZSBkZWxldGVkIGFmdGVyIGEgc3VjY2Vzc2Z1bCB1cGxvYW'
    b'QsIGFuZCBtb3ZlZCB0byBhIHN1YmZvbGRlciBsYWJlbGVkICdmYWlsZWQnIGluIHRoZSBj'
    b'YXNlIG9mIGEgZmFpbGVkIHVwbG9hZEAnQMJQAYMd/39RAYEAQQepUAWmS2VlcCB0aGUgV2'
    b'ktRmkgbW9kdWxlIG9uIGFuZCBjb25uZWN0ZWRQFUCISWYgY2hlY2tlZCwgdGhlIFdpLUZp'
    b'IG1vZHVsZSB3aWxsIHN0YXkgb24gYW5kIGNvbm5lY3RlZC4gIEFsbG93cyBmb3IgZmFzdG'
    b'VyIHVwbG9hZCB0aW1lcyB3aXRoIHRoZSBleHBlbnNlIG9mIGluY3JlYXNlZCBwb3dlciBj'
    b'b25zdW1wdGlvbkAPgEBPgEAoSOhQBYhUcmlnZ2Vyc0AHQNpQBYxUcmlnZ2VyIE1vZGVQAY'
    b'MS/39RAYEAUBVAj0RlbGF5IFRoZW4gVHJpZ2dlcjogV2FpdCBzcGVjaWZpZWQgdGltZSwg'
    b'dGhlbiB3YWl0IGZvciB0cmlnZ2VyCkRlbGF5IE9yIFRyaWdnZXI6IEFjcXVpcmUgYWZ0ZX'
    b'Igd2FpdCBvciBvbiB0cmlnZ2VyLiBCYXR0ZXJ5IGxpZmUgbWF5IGJlIHJlZHVjZWQuQQeV'
    b'UAWSRGVsYXkgVGhlbiBUcmlnZ2VyQQeTUAWQRGVsYXkgT3IgVHJpZ2dlckAy9FAFjVN0YX'
    b'J0IGF0IFRpbWVQAYMP/39QFdtXYWl0IHVudGlsIHNldCB0aW1lIGJlZm9yZSBldmFsdWF0'
    b'aW5nIG90aGVyIHRyaWdnZXIgY29uZGl0aW9ucywgaW5jbHVkaW5nIGEgcmVjb3JkaW5nIG'
    b'RlbGF5QBGuUAWUUmVjb3JkaW5nIFRpbWUgTGltaXRQAYMN/39RIYT////wUCWHU2Vjb25k'
    b'c0ARQJBQBZhSZWNvcmRpbmcgRGVsYXkvSW50ZXJ2YWxQAYMM/39RAYEAUCWHU2Vjb25kc1'
    b'EhgwFRgFAV2FRpbWUgYmVmb3JlIHN0YXJ0aW5nIGEgcmVjb3JkaW5nLCBhbmQgdGltZSBi'
    b'ZXR3ZWVuIHJlY29yZGluZ3MgaWYgJ1JldHJpZ2dlcicgaXMgY2hlY2tlZC5AEa9QBZlSZW'
    b'NvcmRpbmcgRmlsZSBTaXplIExpbWl0UAGDEf9/USGE////8FAlg0tpQkAQQJdQBYlSZXRy'
    b'aWdnZXJQAYMO/39QFLBDb25maWdbMHhERkY3Rl09PW51bGwgYW5kIENvbmZpZ1sweDExRk'
    b'Y3Rl09PW51bGxQFc9XaGVuIHNldCwgU2xhbSBTdGljayB3aWxsIHJlLWFybSBhZnRlciBy'
    b'ZWNvcmRpbmcgdGltZSBvciBzaXplIGxpbWl0IGlzIHJlYWNoZWQuQBBAqFAFnldhaXQgZm'
    b'9yIEFsbCBTZW5zb3IgQ29uZGl0aW9uc1ABgxP/f1AV/ldoZW4gc2V0LCByZWNvcmRpbmcg'
    b'd2lsbCBvbmx5IHN0YXJ0IHdoZW4gYWxsIHRyaWdnZXJzIGFyZSBtZXQsIG90aGVyd2lzZS'
    b'ByZWNvcmRpbmcgd2lsbCBzdGFydCB3aGVuIGFueSBzZW5zb3IgY29uZGl0aW9uIGlzIG1l'
    b'dEAYQO5QBZBQcmVzc3VyZSBUcmlnZ2VyUAGDBQAkUQCBAECC5VAFlVByZXNzdXJlIFRyaW'
    b'dnZXIsIExvd1ABgwMAJFEBgwFfkFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0YXJ0IHNhbXBs'
    b'aW5nIHdoZW4gb3V0c2lkZSBwcmVzc3VyZSB3aW5kb3cuQILmUAWWUHJlc3N1cmUgVHJpZ2'
    b'dlciwgSGlnaFABgwQAJFEBgwGtsFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0YXJ0IHNhbXBs'
    b'aW5nIHdoZW4gb3V0c2lkZSBwcmVzc3VyZSB3aW5kb3cuQBhA+1AFk1RlbXBlcmF0dXJlIF'
    b'RyaWdnZXJQAYMFASRRAIEAQGPqUAWYVGVtcGVyYXR1cmUgVHJpZ2dlciwgTG93UAGDAwEk'
    b'UQKB9lAlgUNREoHYUSKBUFAVtlNldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZG'
    b'UgdGVtcGVyYXR1cmUgd2luZG93LkBj61AFmVRlbXBlcmF0dXJlIFRyaWdnZXIsIEhpZ2hQ'
    b'AYMEASRRAoEjUCWBQ1ESgdhRIoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3'
    b'V0c2lkZSB0ZW1wZXJhdHVyZSB3aW5kb3cuQBhBqFAFm0FjY2VsZXJhdGlvbiBUcmlnZ2Vy'
    b'ICgxMDBnKVABgwX/CFARgwf/IFEAgQBQJIBQFJVDb25maWdbMHgwMUZGMDhdID09IDBAQ0'
    b'CrUROIwFkAAAAAAABRI4hAWQAAAAAAAFAFmUFjY2VsZXJhdGlvbiBUcmlnZ2VyLCBMb3dQ'
    b'AYMD/whRAoL5mlAV4EhpZ2ggcG93ZXIgdHJpZ2dlciB3aXRoIHByZS10cmlnZ2VyIGRhdG'
    b'EuIE5vdCByZWNvbW1lbmRlZCBmb3IgbG9uZyB3YWl0cyB3aXRob3V0IGV4dGVybmFsIHBv'
    b'd2VyLlIDiD9pABkAGQAZQENArFETiMBZAAAAAAAAUSOIQFkAAAAAAABQBZpBY2NlbGVyYX'
    b'Rpb24gVHJpZ2dlciwgSGlnaFABgwT/CFECggZmUBXgSGlnaCBwb3dlciB0cmlnZ2VyIHdp'
    b'dGggcHJlLXRyaWdnZXIgZGF0YS4gTm90IHJlY29tbWVuZGVkIGZvciBsb25nIHdhaXRzIH'
    b'dpdGhvdXQgZXh0ZXJuYWwgcG93ZXIuUgOIP2kAGQAZABlAGEGjUAWaQWNjZWxlcmF0aW9u'
    b'IFRyaWdnZXIgKDQwZylQAYMH/1BQFdRMb3cgcG93ZXIgdHJpZ2dlciB1c2luZyBkaWdpdG'
    b'FsIGFjY2VsZXJvbWV0ZXIgd2l0aCBEQyByZXNwb25zZS4gTm8gcHJlLXRyaWdnZXIgZGF0'
    b'YS5RAIEAUBGDBf8IUBGDB/8gUCSAUBSUQ29uZmlnWzB4MWZmNTBdID09IDBAQ0CyUAWWQW'
    b'NjZWxlcmF0aW9uIFRocmVzaG9sZFABgwT/UFEBgiAAUgOIP0QAFAAUABRRI4hARAAAAAAA'
    b'AFETgFAV8kxvdyBwb3dlciB0cmlnZ2VyLiBUaGUgbWluaW11bSBhY2NlbGVyYXRpb24gKH'
    b'Bvc2l0aXZlIG9yIG5lZ2F0aXZlKSB0byB0cmlnZ2VyIGEgcmVjb3JkaW5nLiBNdXN0IGJl'
    b'IGdyZWF0ZXIgdGhhbiAwLkAnxlABgwX/UFEBgQdBB5FQBY5YIEF4aXMgVHJpZ2dlckEHkV'
    b'AFjlkgQXhpcyBUcmlnZ2VyQQeRUAWOWiBBeGlzIFRyaWdnZXJAD4BAT4BAKEfjUAWMTWVh'
    b'c3VyZW1lbnRzQAhCW1AFmDEwMGcgQWNjZWxlcmF0aW9uIChDaCA4KVAkgEAnQJtRAYEPUA'
    b'GDAf8IUBWgRW5hYmxlL0Rpc2FibGUgYWNjZWxlcmF0aW9uIGF4ZXNBB5hQBZFFbmFibGUg'
    b'WCAoQ2ggOC4wKVEBgQBBB5hQBZFFbmFibGUgWSAoQ2ggOC4xKVEBgQFBB5hQBZFFbmFibG'
    b'UgWiAoQ2ggOC4yKVEBgQJBB5pQBZNFbmFibGUgTWljIChDaCA4LjMpUQGBA0ARQJpQBYtT'
    b'YW1wbGUgUmF0ZVABgwL/CFAV2FNldCBtYWluIGFjY2VsZXJhdGlvbiBzYW1wbGUgcmF0ZS'
    b'Bmcm9tIDEwIHRvIDIwMDAwIEh6LgpIaWdoZXIgcmF0ZSByZWR1Y2VzIGJhdHRlcnkgbGlm'
    b'ZS5QFJVDb25maWdbMHgwMUZGMDhdID09IDBRAYITiFAlgkh6URGBClEhgk4gQBFA/FAFo0'
    b'92ZXJyaWRlIEFudGlhbGlhc2luZyBGaWx0ZXIgQ3V0b2ZmUAGDCP8IUQGCA+hQJYJIelEh'
    b'gmGoUBVApURlZmF1bHQsIHdoZW4gbm90IGNoZWNrZWQsIG9mIG9uZS1maWZ0aCBzYW1wbG'
    b'UgZnJlcXVlbmN5LiBDdXQtb2ZmIGZyZXF1ZW5jeSBoYXMgMjklIGF0dGVudWF0aW9uLgo1'
    b'dGggb3JkZXIgQnV0dGVyd29ydGguIFNvbWUgYXR0ZW51YXRpb24gYXQgNjAlIG9mIGN1dC'
    b'1vZmYgZnJlcXVlbmN5LlAUlUNvbmZpZ1sweDAxRkYwOF0gPT0gMEAIQRxQBZs0MGcgREMg'
    b'QWNjZWxlcmF0aW9uIChDaCA4MClAJ59QAYMB/1BRAYEBQQeSUAWPRW5hYmxlIEFsbCBBeG'
    b'VzQBdA2FAFi1NhbXBsZSBSYXRlUAGDAv9QUQGCAfRQJYJIelAV3kxvdyBwYXNzIGZpbHRl'
    b'ciBpcyBzZXQgdG8gMjUlIG9mIGRhdGEgcmF0ZSwgc28gYSBkYXRhIHJhdGUgb2YgMjAwME'
    b'h6IGhhcyBhIExQIGZpbHRlciBhdCA1MDAgSHpQFJNub3QgQ29uZmlnWzB4MWZmNTBdQQeF'
    b'UQGCD6BBB4VRAYIH0EEHhVEBggPoQQeFUQGCAfRBB4RRAYH6QQeEUQGBfUEHhFEBgT9BB4'
    b'RRAYEgQQeEUQGBEEAIQTBQBahDb250cm9sIFBhZCBUZW1wZXJhdHVyZS9QcmVzc3VyZSAo'
    b'Q2ggNTkpQCdAm1EBgQNQAYMB/ztQFaVFbmFibGUvRGlzYWJsZSBzZW5zb3JzIG9uIGNvbn'
    b'Ryb2wgcGFkQQecUAWZRW5hYmxlIFByZXNzdXJlIChDaCA1OS4wKUEHn1AFnEVuYWJsZSBU'
    b'ZW1wZXJhdHVyZSAoQ2ggNTkuMSlBB6VQBaJFbmFibGUgUmVsYXRpdmUgSHVtaWRpdHkgKE'
    b'NoIDU5LjIpQBHjUAWLU2FtcGxlIFJhdGVQAYMC/ztQFJRDb25maWdbMHgxZmYzQl0gPT0g'
    b'MFAVpFNldCBzYW1wbGUgZnJlcXVlbmN5IGZyb20gMSB0byAxMCBIelEBgQpQJYJIelERgQ'
    b'FRIYEKQAhBzVAFmUluZXJ0aWFsIE1lYXN1cmVtZW50IFVuaXRAB0CWUAWQQWNxdWlzaXRp'
    b'b24gTW9kZVABgwH/K1AlgSBRAYEQQQeKUAWDT2ZmUQGBAEEHqFAFoUFic29sdXRlIE9yaW'
    b'VudGF0aW9uIChRdWF0ZXJuaW9uKVEBgQhBB6hQBaFSZWxhdGl2ZSBPcmllbnRhdGlvbiAo'
    b'UXVhdGVybmlvbilRAYEQQQePUAWIUm90YXRpb25RAYECQAFAhlAFlU9yaWVudGF0aW9uIE'
    b'RhdGEgUmF0ZVABgwIIK1EBgWRQJYJIelERgQFRIYHIUBW9U2FtcGxlIHJhdGUgZm9yIGRp'
    b'cmVjdGlvbmFsIGRhdGEgZnJvbSAxIHRvIDIwMCBIei4gQ2hhbm5lbCA2NVAUlENvbmZpZ1'
    b'sweDAxRkYyQl0gPCA4QAFAiVAFklJvdGF0aW9uIERhdGEgUmF0ZVABgwICK1EBgWRQJYJI'
    b'elERgQFRIYHIUBW+U2V0IHNhbXBsZSByYXRlIGZvciByb3RhdGlvbiBkYXRhIGZyb20gMS'
    b'B0byAyMDAgSHouIENoYW5uZWwgNDdQFJlDb25maWdbMHgwMUZGMkJdICYgMiA9PSAwQAhA'
    b'kFAFoENvbnRyb2wgUGFkIExpZ2h0IFNlbnNvciAoQ2ggNzYpQCfqUAGDAf9MUBW+RW5hYm'
    b'xlL0Rpc2FibGUgbGlnaHQgc2Vuc29yIGF0IDQgSHogLSBOT1RFOiBESVNBQkxFRCBGT1Ig'
    b'QUxQSEFBB6BQBZ1FbmFibGUgTGlnaHQgU2Vuc29yIChDaCA3Ni4wKUAIQLJQBZNCTUcyNT'
    b'AgR3lybyAoQ2ggODQpQCefUAGDAf9UUQGBAEEHklAFj0VuYWJsZSBBbGwgQXhlc0AX91AF'
    b'i1NhbXBsZSBSYXRlUAGDAv9UUQGCAZBQJYJIelAUk25vdCBDb25maWdbMHgxZmY1NF1BB4'
    b'VRAYIMgEEHhVEBggZAQQeFUQGCAyBBB4VRAYIBkEEHhFEBgchBB4RRAYFkQQeEUQGBMkEH'
    b'hFEBgRlBB4RRAYEMQA+AQE+ASgiASiiASkiA'
)
